from __future__ import annotations
from dataclasses import dataclass
import rsyscall.far
import rsyscall.near
import typing as t
import logging
import contextlib
from rsyscall.struct import Serializer
from rsyscall.memory.allocation_interface import AllocationInterface, UseAfterFreeError
from rsyscall.memory.transport import MemoryGateway
from rsyscall.sys.mman import MemoryMapping
logger = logging.getLogger(__name__)


T = t.TypeVar('T')
T_co = t.TypeVar('T_co', covariant=True)
U = t.TypeVar('U')
T_pointer = t.TypeVar('T_pointer', bound='Pointer')
@dataclass(eq=False)
class Pointer(t.Generic[T]):
    """An owning handle for some piece of memory.

    More precisely, this is an owning handle for an allocation in some memory mapping.  We're
    explicitly representing memory mappings, rather than glossing over them and pretending that the
    address space is flat and uniform. If we have two mappings for the same file, we can translate
    this Pointer between them.

    As an implication of owning an allocation, we also know the length of that allocation, which is
    the length of the range of memory that it's valid to operate on through this pointer. We
    retrieve this through Pointer.size and use it in many places; anywhere we take a Pointer, if
    there's some question about what size to operate on, we operate on the full size of the
    pointer. Reducing the amount of memory to operate on can be done through Pointer.split.

    We also know the type of the region of memory; that is, how to interpret this region of
    memory. This is useful at type-checking time to check that we aren't passing pointers to memory
    of the wrong type. At runtime, the type is reified as a serializer, which allows us to translate
    a value of the type to and from bytes.

    We also hold a transport which will allow us to read and write the memory we own. Combined with
    the serializer, this allows us to write and read values of the appropriate type to and from
    memory using the Pointer.write and Pointer.read methods.

    Finally, pointers have a "valid" bit which says whether the Pointer can be used. We say that a
    method "consumes" a pointer if it will invalidate that pointer.

    Most of the methods manipulating the pointer are "linear". That is, they consume the pointer
    object they're called on and return a new pointer object to use. This forces the user to be more
    careful with tracking the state of the pointer; and also allows us to represent some state
    changes with by changing the type of the pointer, in particular Pointer.write.

    See also the inheriting class WrittenPointer

    """
    __slots__ = ('mapping', 'transport', 'serializer', 'allocation', 'valid', 'typ')
    mapping: MemoryMapping
    transport: MemoryGateway
    serializer: Serializer[T]
    allocation: AllocationInterface
    typ: t.Type[T]
    valid: bool

    def __init__(self,
                 mapping: MemoryMapping,
                 transport: MemoryGateway,
                 serializer: Serializer[T],
                 allocation: AllocationInterface,
                 typ: t.Type[T],
    ) -> None:
        self.mapping = mapping
        self.transport = transport
        self.serializer = serializer
        self.allocation = allocation
        self.typ = typ
        self.valid = True

    async def write(self, value: T) -> WrittenPointer[T]:
        "Write this value to this pointer, consuming it and returning a new WrittenPointer"
        self._validate()
        value_bytes = self.serializer.to_bytes(value)
        if len(value_bytes) > self.size():
            raise Exception("value_bytes is too long", len(value_bytes),
                            "for this typed pointer of size", self.size())
        await self.transport.write(self, value_bytes)
        return self._wrote(value)

    async def read(self) -> T:
        "Read the value pointed to by this pointer"
        self._validate()
        value = await self.transport.read(self)
        return self.serializer.from_bytes(value)

    def size(self) -> int:
        """Return the size of this pointer's allocation in bytes

        This is mostly used by syscalls and passed to the kernel, so that the kernel knows the size
        of the buffer that it's been passed. To reduce the size of a buffer passed to the kernel,
        use Pointer.split.

        """
        return self.allocation.size()

    def split(self, size: int) -> t.Tuple[Pointer, Pointer]:
        """Invalidate this pointer and split it into two adjacent pointers

        This is primarily used by syscalls that write to one contiguous part of a buffer and leave
        the rest unused.  They split the pointer into a "used" part and an "unused" part, and return
        both parts.

        """
        self._validate()
        # TODO uhhhh if split throws an exception... don't we need to free... or something...
        self.valid = False
        # TODO we should only allow split if we are the only reference to this allocation
        alloc1, alloc2 = self.allocation.split(size)
        first = self._with_alloc(alloc1)
        # TODO should degrade this pointer to raw bytes or something, or maybe no type at all
        second = self._with_alloc(alloc2)
        return first, second

    def merge(self, ptr: Pointer) -> Pointer:
        """Merge two pointers produced by split back into a single pointer

        The two pointers passed in are invalidated.

        This is primarily used by the user to re-assemble a buffer that was split by a syscall.

        """
        self._validate()
        ptr._validate()
        # TODO should assert that these two pointers both serialize the same thing
        # although they could be different types of serializers...
        self.valid = False
        ptr.valid = False
        # TODO we should only allow merge if we are the only reference to this allocation
        alloc = self.allocation.merge(ptr.allocation)
        return self._with_alloc(alloc)

    def __add__(self, right: Pointer[T]) -> Pointer[T]:
        "left + right desugars to left.merge(right)"
        return self.merge(right)

    def __radd__(self, left: t.Optional[Pointer[T]]) -> Pointer[T]:
        """"left += right" desugars to "left = (left + right) if left is not None else right"

        With this, you can initialize a variable to None, then merge pointers into it in a
        loop. This is especially useful when trying to write an entire buffer, or fill an
        entire buffer by reading.

        """
        if left is None:
            return self
        else:
            return left + self

    @property
    def near(self) -> rsyscall.near.Address:
        """Return the raw memory address referred to by this Pointer

        This is mostly used by syscalls and passed to the kernel, so that the kernel knows the start
        of the buffer to read to or write from.

        """
        # TODO hmm should maybe validate that this fits in the bounds of the mapping I guess
        self._validate()
        try:
            return self.mapping.near.as_address() + self.allocation.offset()
        except UseAfterFreeError as e:
            raise UseAfterFreeError(
                "Allocation inside this Pointer", self,
                "is freed, but the pointer is still valid; someone violated some invariants",
            ) from e

    def check_address_space(self, task: rsyscall.far.Task) -> None:
        if task.address_space != self.mapping.task.address_space:
            raise rsyscall.far.AddressSpaceMismatchError(task.address_space, self.mapping.task.address_space)

    @contextlib.contextmanager
    def borrow(self, task: rsyscall.far.Task) -> t.Iterator[rsyscall.near.Address]:
        """Pin the address of this pointer, and yield the pointer's raw memory address

        We validate this pointer, and pin it in memory so that it can't be moved or deleted while
        it's being used.

        This is mostly used by syscalls and passed to the kernel, so that the kernel knows the start
        of the buffer to read to or write from.

        """
        # TODO actual tracking of pointer references is not yet implemented
        # we should have a flag or lock to indicate that this pointer shouldn't be moved or deleted,
        # while it's being borrowed.
        # TODO rename this to pinned
        # TODO make this the only way to get .near
        self._validate()
        self.check_address_space(task)
        yield self.near

    def _validate(self) -> None:
        if not self.valid:
            raise UseAfterFreeError("handle is no longer valid")

    def free(self) -> None:
        """Free this pointer, invalidating it and releasing the underlying allocation.

        It isn't necessary to explicitly call this, because the pointer will be freed on
        GC. But you can call it anyway if, for example, the pointer will be referenced for
        long after it is done being used.

        """
        if self.valid:
            self.valid = False
            self.allocation.free()

    def __del__(self) -> None:
        # This isn't strictly necessary because the allocation will free itself on __del__.
        # But, that will only happen when *all* pointers referring to the allocation are collected;
        # not just the valid one.
        # So, this ensures GC is a bit more prompt.
        # Oh, wait. The real reason we need this is because the Arena stores references to the allocation.
        # TODO We should fix that.
        self.free()

    def split_from_end(self, size: int, alignment: int) -> t.Tuple[Pointer, Pointer]:
        """Split from the end of this pointer, such that the right pointer is aligned to `alignment`

        Used by write_to_end; mostly only useful for preparing stacks.

        """
        extra_to_remove = (int(self.near) + size) % alignment
        return self.split(self.size() - size - extra_to_remove)

    async def write_to_end(self, value: T, alignment: int) -> t.Tuple[Pointer[T], WrittenPointer[T]]:
        """Write a value to the end of the range of this pointer

        Splits the pointer, and returns both parts.  This function is only useful for preparing
        stacks. Would be nice to figure out either a more generic way to prep stacks, or to figure
        out more things that write_to_end could be used for.

        """
        value_bytes = self.serializer.to_bytes(value)
        rest, write_buf = self.split_from_end(len(value_bytes), alignment)
        written = await write_buf.write(value)
        return rest, written

    def __repr__(self) -> str:
        name = type(self).__name__
        typname = self.typ.__name__
        try:
            return f"{name}[{typname}]({self.near}, {self.size()})"
        except UseAfterFreeError:
            return f"{name}[{typname}](valid={self.valid}, {self.mapping}, {self.allocation}, {self.serializer})"

    #### Various ways to create new Pointers by changing one thing about the old pointer. 
    def _with_mapping(self: T_pointer, mapping: MemoryMapping) -> T_pointer:
        if type(self) is not Pointer:
            raise Exception("subclasses of Pointer must override _with_mapping")
        if mapping.file is not self.mapping.file:
            raise Exception("can only move pointer between two mappings of the same file")
        # we don't have a clean model for referring to the same object through multiple mappings.
        # this is a major TODO.
        # at least two ways to achieve it:
        # - have Pointers become multi-mapping super-pointers, which can be valid in multiple address spaces
        # - break our linearity constraint on pointers, allowing multiple pointers for the same allocation;
        #   this is difficult because split() is only easy to implement due to linearity.
        # right here, we just linearly move the pointer to a new mapping
        self._validate()
        self.valid = False
        return type(self)(mapping, self.transport, self.serializer, self.allocation, self.typ)

    def _with_alloc(self, allocation: AllocationInterface) -> Pointer:
        return Pointer(self.mapping, self.transport, self.serializer, allocation, self.typ)

    def _reinterpret(self, serializer: Serializer[U], typ: t.Type[U]) -> Pointer[U]:
        # TODO how can we check to make sure we don't reinterpret in wacky ways?
        # maybe we should only be able to reinterpret in ways that are allowed by the serializer?
        # so maybe it's a method on the Serializer? cast_to(Type)?
        self._validate()
        self.valid = False
        return Pointer(self.mapping, self.transport, serializer, self.allocation, typ)

    def _readable(self) -> ReadablePointer[T]:
        self._validate()
        self.valid = False
        return ReadablePointer(self.mapping, self.transport, self.serializer, self.allocation, self.typ)

    def readable_split(self, size: int) -> t.Tuple[ReadablePointer[T], Pointer]:
        left, right = self.split(size)
        return left._readable(), right

    def _linearize(self) -> LinearPointer[T]:
        self._validate()
        self.valid = False
        return LinearPointer(self.mapping, self.transport, self.serializer, self.allocation, self.typ)

    def unsafe(self) -> ReadablePointer[T]:
        "Get a ReadablePointer from this pointer, even though it might not be initialized"
        return self._readable()

    def _wrote(self, value: T) -> WrittenPointer[T]:
        "Assert we wrote this value to this pointer, and return the appropriate new WrittenPointer"
        self._validate()
        self.valid = False
        return WrittenPointer(self.mapping, self.transport, value, self.serializer, self.allocation, self.typ)

class ReadablePointer(Pointer[T]):
    """A Pointer that is safely readable

    This is returned by functions and syscalls which write some (possibly
    unknown) pure data to an address in memory, which then can be read and
    deserialized to get a sensical pure data value rather than nonsense.

    Immediately after allocation, a Pointer is returned, rather than a
    ReadablePointer, to indicate that the pointer is uninitialized, and
    therefore not safely readable.

    This is also returned by Pointer.unsafe(), to support system calls where
    it's not statically known that a passed Pointer is written to and
    initialized; ioctls, for example.  Tt would be better to have a complete
    description of the Linux interface, so we could get rid of this unsafety.

    This is currently only a marker type, but eventually we'll move the read()
    method here to ReadablePointer from Pointer, so that reading Pointers is
    actually not allowed. For now, this is just a hint.

    """
    __slots__ = ()

    def _with_mapping(self, mapping: MemoryMapping) -> ReadablePointer:
        # see notes in Pointer._with_mapping
        if type(self) is not ReadablePointer:
            raise Exception("subclasses of ReadablePointer must override _with_mapping")
        if mapping.file is not self.mapping.file:
            raise Exception("can only move pointer between two mappings of the same file")
        self._validate()
        self.valid = False
        return type(self)(mapping, self.transport, self.serializer, self.allocation, self.typ)

class LinearPointer(ReadablePointer[T]):
    """A Pointer that must be read, once

    This is returned by functions and syscalls which write a unknown
    value to an address in memory, which then must be read and
    deserialized *once* to manage the resources described by that
    value, such as file descriptors.

    The value is:

    - "affine"; it must be read at least once, so that the resources inside
      can be returned as managed objects.
    - "relevant"; it must be read at most once, so that dangling handles to the
      resources can't be created again after they're closed.

    Since it's both affine and relevant, this is a true linear type.

    Unfortunately it's going to be quite difficult to guarantee relevance. There
    are three issues here:

    1. The pointer can simply be dropped and garbage collected.
    2. System calls can write to the pointer and discard its previous results
    3. We can write to the pointer (through `Pointer.write`) and discard its previous results

    We can mitigate 1 a little by warning in `__del__`.

    We could statically prevent 3 by removing the `Pointer.read` and `Pointer.write` methods
    from this class, and only allowing `LinearPointer.linear_read`, or dynamically by throwing
    in `write` if `been_read` is false.

    Any approach to 2 is going to require some tweaks to the pointer API, and
    probably some mass changes to syscall implementations. Although maybe we
    could do it off of .near accesses.

    """
    __slots__ = ('been_read')

    def __init__(self,
                 mapping: MemoryMapping,
                 transport: MemoryGateway,
                 serializer: Serializer[T],
                 allocation: AllocationInterface,
                 typ: t.Type[T],
    ) -> None:
        super().__init__(mapping, transport, serializer, allocation, typ)
        self.been_read = False

    async def read(self) -> T:
        if self.been_read:
            raise Exception("This LinearPointer has already been read, it can't be read again for safety reasons.")
        ret = await super().read()
        self.been_read = True
        return ret

    async def linear_read(self) -> t.Tuple[T, Pointer[T]]:
        "Read the value, and return the now-inert buffer left over as a Pointer."
        ret = await self.read()
        self.valid = False
        new_ptr = Pointer(self.mapping, self.transport, self.serializer, self.allocation, self.typ)
        return ret, new_ptr

    def __del__(self) -> None:
        super().__del__()
        if not self.been_read:
            logger.error("Didn't read this LinearPointer before dropping it: %s", self)

    def _with_mapping(self, mapping: MemoryMapping) -> LinearPointer:
        # see notes in Pointer._with_mapping
        if type(self) is not LinearPointer:
            raise Exception("subclasses of LinearPointer must override _with_mapping")
        if mapping.file is not self.mapping.file:
            raise Exception("can only move pointer between two mappings of the same file")
        self._validate()
        self.valid = False
        return type(self)(mapping, self.transport, self.serializer, self.allocation, self.typ)

class WrittenPointer(Pointer[T_co]):
    """A Pointer with some known value written to it

    We have all the normal functionality of a Pointer (see that class for more information), but we
    also know that we've had some value written to us, and we know what that value is, and it's
    immediately accessible in Python.

    We can also view this with an emphasis on the value: This is some known value, that has been
    written to some memory location. The value and the pointer are equally important in this class,
    and both are used by most uses of this class.

    We use inheritance so that a WrittenPointer gracefully degrades back to a Pointer, and is
    invalidated whenever a pointer is invalidated. Specifically, we want anything that writes to a
    pointer to invalidate this pointer. The invalidation lets us know that this value is no longer
    necessarily written to this pointer.

    For example, syscalls that write to pointers will typically call split. A call to
    WrittenPointer.split will invalidate the WrittenPointer and return regular Pointers; that's
    desirable because the syscall likely overwrote whatever value was previously written here.

    TODO: We should fix syscalls that write to memory but don't call split so that they invalidate
    the WrittenPointer. That's mostly syscalls using Sockbufs...

    """
    __slots__ = ('value')
    def __init__(self,
                 mapping: MemoryMapping,
                 transport: MemoryGateway,
                 value: T_co,
                 serializer: Serializer[T_co],
                 allocation: AllocationInterface,
                 typ: t.Type[T_co],
    ) -> None:
        super().__init__(mapping, transport, serializer, allocation, typ)
        self.value = value

    def __repr__(self) -> str:
        name = type(self).__name__
        typname = self.typ.__name__
        try:
            return f"{name}[{typname}]({self.near}, {self.value})"
        except UseAfterFreeError:
            return f"{name}[{typname}](valid={self.valid}, {self.mapping}, {self.allocation}, {self.value})"

    def _with_mapping(self, mapping: MemoryMapping) -> WrittenPointer:
        if type(self) is not WrittenPointer:
            raise Exception("subclasses of WrittenPointer must override _with_mapping")
        if mapping.file is not self.mapping.file:
            raise Exception("can only move pointer between two mappings of the same file")
        # see notes in Pointer._with_mapping
        self._validate()
        self.valid = False
        return type(self)(mapping, self.transport, self.value, self.serializer, self.allocation, self.typ)



"""Non-blocking IO operations implemented using epoll.

We use epoll in edge-triggered mode. Our usage is relatively
straightforward, but there are a few quirks.


--------------------------------------------------------------------------------

To avoid blocking a thread in a call to epoll_wait when there is more
work to be done, there are two options:

1. We can register the epollfd on some higher event loop, and instead
of blocking in epoll_wait, just wait until that event loop tells us
there are events to be read. This is the typical way to use most file
descriptors in a non-blocking way, just applied to an epollfd.

2. Or, we can ensure that all possible pending work is registered on
our epollfd; anything that might want to do some work needs to
register a file descriptor that is readable/writable/etc on our
epollfd.  This is the typical requirement that most event loops place
on the rest of the program.

In an ideal world, we would do both with every epoll instance.
Whenever a new epoll E is created, we would register E on all the
other event loops; and we would register all the other event loops on
E. Then we could freely block in either E or any other event loop, and
not miss any events. We could have multiple event loops in our process
without pain.

Unfortunately, this would necessitate a cycle in epoll registrations
(i.e. an epoll A registered on an epoll B which is itself registered
on epoll A). This would necessitate breaking the cycle each time
epoll_wait is called, which could be non-trivially expensive, and
Linux doesn't want to do that.

So instead, to prevent cycles, we need to use only one of these two
approaches at any one time; at any one time, we'll have a single
"root" event loop which uses approach 2 and can make blocking calls to
epoll_wait; and the other event loops will be "subsidiary", using
approach 1, and delegating their blocking to the root event loop.
Then we can still have multiple event loops in our process.

These two approaches to making an Epoller are implemented in the
"make_subsidiary" (1) and "make_root" (2) classmethods of Epoller.

We would rather go with 2 - the only one allowed to block in our
threads should be us. Unfortunately, most event loop libraries don't
expose a way to function as a subsidiary event loop.  So in the case
of threads which need to interoperate with another event loop, we are
forced into 1.

In other cases, there is no other event loop, so we must go with 2.
We must take care to satisfy the requirements of 2: Anything that
wants to perform work should cause a wakeup on our epollfd.

The only case where we use 1 is when running on the local Python
thread.  There, we use the trio async library; and trio does not
expose its epollfd such that we could use approach 2. Instead, we call
trio's wait_readable method on the epollfd instead of blocking in
epoll_wait.

On every thread other than the local Python thread, we use approach
2. At the moment, the only other source of work that is present in
every thread is the file descriptor from which the rsyscall server
reads incoming syscalls. So that other syscalls will not be blocked by
our calls to epoll_wait, we register that file descriptor (exposed in
SyscallInterface as the "activity fd") on our epollfd. In this way, if
a new syscall is received in the thread while already in a call to
epoll_wait, the epoll_wait will wake up so that the rsyscall server
can read and perform the unrelated syscalls.

We could straightforwardly support a mode for an Epoller to be
registered on another of our Epollers; but we haven't done this
because it's not really useful.


--------------------------------------------------------------------------------

"""
from __future__ import annotations
from dneio import RequestQueue, Continuation, reset
from rsyscall._raw import ffi # type: ignore
import collections
import errno
import os
import math
import typing as t
from rsyscall.near.sysif import SyscallError
from rsyscall.memory.ram import RAM
from rsyscall.handle import FileDescriptor, Pointer, WrittenPointer, Task
from dataclasses import dataclass
from rsyscall.struct import Int32, T_fixed_size
from rsyscall.sys.syscall import SYS
from rsyscall.sys.socket import SOCK, SOL, SO, Sockaddr, SockaddrStorage, T_sockaddr, Sockbuf
from rsyscall.sys.epoll import EpollEvent, EpollEventList, EPOLL, EPOLL_CTL, EpollFlag
from rsyscall.fcntl import O, F

import logging
logger = logging.getLogger(__name__)

__all__ = [
    "Epoller",
    "EpolledFileDescriptor",
    "AsyncFileDescriptor",
    "AsyncReadBuffer",
]

class RemovedFromEpollError(Exception):
    pass

class EpollWaiter:
    """The core class which reads events from the epollfd and dispatches them.

    Many threads may have a copy of our epollfd, and register fds on it. To actually wait
    on this epollfd and pull events from it is the responsibility of this class. By
    inheriting the epollfd to many threads for registration purposes, but centralizing the
    actual epoll_wait calls in this shared class, our usage of epoll becomes more
    efficient.

    """
    def __init__(self, ram: RAM, epfd: FileDescriptor,
                 wait_readable: t.Optional[t.Callable[[], t.Awaitable[None]]],
                 timeout: int,
    ) -> None:
        "To make this, use one of the constructor methods of Epoller: make_subsidiary or make_root"
        self.ram = ram
        self.epfd = epfd
        self.wait_readable = wait_readable
        self.timeout = timeout

        self.used_numbers: t.Set[int] = set()
        self.pending_remove: t.Set[int] = set()
        self.queue = RequestQueue[int, EPOLL]()
        reset(self._run())

    def allocate_number(self, request: int) -> int:
        """Add a callback which will be called on EpollEvents with data == returned number.

        We can then add the returned number to the epollfd with some file descriptor and
        mask.

        This is not the user-interface for epoll. That's the register method on Epoller.

        """
        number = request
        # this is amusingly horrible, but it provides a useful feature:
        # the number in epoll usually matches the fd number.
        while number in self.used_numbers:
            number += 1000
        self.used_numbers.add(number)
        return number

    def remove_number_after_next_epoll_wait(self, number: int) -> None:
        """Remove the callback for this number after the next time we call epoll_wait.

        We don't want to remove it immediately, since even if this
        number was just removed, we might still get an event for it if
        an epoll_wait call was in progress.
        """
        self.pending_remove.add(number)

    async def _run(self) -> None:
        input_buf: Pointer = await self.ram.malloc(EpollEventList, 32 * EpollEvent.sizeof())
        number_to_cb: t.Dict[int, Continuation[EPOLL]] = {}
        registered_activity_fd: t.Optional[FileDescriptor] = None
        while True:
            if self.wait_readable:
                await self.wait_readable()
            activity_fd = self.epfd.task.sysif.get_activity_fd()
            if activity_fd and (registered_activity_fd is not activity_fd):
                # the activity fd changed, we need to register the new one
                if registered_activity_fd:
                    # delete the old registered activity fd
                    await self.epfd.epoll_ctl(EPOLL_CTL.DEL, registered_activity_fd)
                activity_fd_number = self.allocate_number(int(activity_fd))
                # start up a coroutine to consume events from the activity_fd
                async def devnull(activity_fd_number=activity_fd_number):
                    while True:
                        await self.queue.request(activity_fd_number)
                reset(devnull())
                await self.epfd.epoll_ctl(EPOLL_CTL.ADD, activity_fd, await self.ram.ptr(
                    EpollEvent(activity_fd_number,
                               # not edge triggered; we don't want to block if there's
                               # anything that can be read.
                               EPOLL.IN|EPOLL.RDHUP|EPOLL.PRI|EPOLL.ERR|EPOLL.HUP)))
                registered_activity_fd = activity_fd
            try:
                valid_events_buf, rest = await self.epfd.epoll_wait(input_buf, self.timeout)
                received_events = await valid_events_buf.read()
            except Exception as wait_error:
                final_exn = wait_error
                break
            input_buf = valid_events_buf + rest
            for num, cb in self.queue.fetch_any():
                number_to_cb[num] = cb
            for event in received_events:
                number_to_cb[event.data].send(event.events)
                del number_to_cb[event.data]
            for num, cb in self.queue.fetch_any():
                number_to_cb[num] = cb
            for number in list(self.pending_remove):
                number_to_cb[number].throw(RemovedFromEpollError())
                del number_to_cb[number]
                self.pending_remove.remove(number)
        self.queue.close(final_exn)

class Epoller:
    "Terribly named class that allows registering fds on epoll, and waiting on them."
    @staticmethod
    def make_subsidiary(ram: RAM, epfd: FileDescriptor, wait_readable: t.Callable[[], t.Awaitable[None]]) -> Epoller:
        """Make a subsidiary epoller, as described in the module docstring.

        We delegate responsibility for blocking to wait for new events to some other
        component. We call the passed-in wait_readable function to block for new events.

        """
        center = Epoller(EpollWaiter(ram, epfd, wait_readable, 0), ram, epfd)
        return center

    @staticmethod
    async def make_root(ram: RAM, task: Task) -> Epoller:
        """Make a root epoller, as described in the module docstring.

        We take responsibility for blocking to wait for new events for every other
        component in this thread. We pull the activity_fd from the SyscallInterface and
        register it on our epollfd. The activity_fd is readable whenever some other
        component in the thread wants to work on the thread.

        """
        epfd = await task.epoll_create()
        center = Epoller(EpollWaiter(ram, epfd, None, -1), ram, epfd)
        return center

    def __init__(self, epoll_waiter: EpollWaiter, ram: RAM, epfd: FileDescriptor) -> None:
        "Don't construct directly; use one of the constructor methods, make_subsidiary or make_root."
        self.epoll_waiter = epoll_waiter
        self.ram = ram
        self.epfd = epfd

    def inherit(self, ram: RAM) -> Epoller:
        """Make a new Epoller which shares the same EpollWaiter class.

        We inherit the epollfd to a new task for the purpose of registering new fds on it;
        but we share the class and task which actually calls epoll_wait.

        """
        return Epoller(self.epoll_waiter, ram, self.epfd.inherit(ram.task))

    async def register(self, fd: FileDescriptor, events: EPOLL) -> EpolledFileDescriptor:
        """Register a file descriptor on this epollfd, for the given events, calling the passed callback.

        The return value can be used to wait for callback calls, modify the events
        registered for this file descriptor, and delete the file descriptor from this
        epollfd.

        """
        number = self.epoll_waiter.allocate_number(int(fd.near))
        efd = EpolledFileDescriptor(self, fd, number)
        await self.epfd.epoll_ctl(EPOLL_CTL.ADD, fd, await self.ram.ptr(EpollEvent(number, events)))
        return efd

class EpolledFileDescriptor:
    """Representation of a file descriptor registered on an epollfd.

    We have to keep around a reference to the file descriptor to perform
    EPOLL_CTL.DEL. This is a bit annoying, but whatever.

    """
    def __init__(self,
                 epoller: Epoller,
                 fd: FileDescriptor,
                 number: int,
    ) -> None:
        self.epoller = epoller
        self.fd = fd
        # TODO, we should copy this so it can't be closed out from under us
        # self.fd = fd.copy()
        self.number = number
        # We optimistically assume that the FD is ready for reading/writing immediately,
        # by setting our initial status to EPOLL.OUT|EPOLL.IN.
        # This has two benefits:
        # 1. Performance improves on our test suite and in many real-world cases, and
        # 2. More critically, if a user erroneously tries to read or write an FD which
        #    will never receive an EPOLL.OUT or EPOLL.IN because it doesn't support
        #    reading/writing, we'll fail immediately instead of blocking forever.
        self.status = FDStatus(EPOLL.OUT|EPOLL.IN)
        self.in_epollfd = True
        self.queue = RequestQueue[EPOLL, None]()
        self.total_events: t.Counter[EPOLL] = collections.Counter()
        self.consumed_events: t.Dict[EPOLL, int] = {flag: 0 for flag in EPOLL}
        reset(self._run())

    def __str__(self) -> str:
        return f"EpolledFileDescriptor({self.fd.near}, {self.number}, {self.status})"

    async def modify(self, events: EPOLL) -> None:
        "Change the EPOLL flags that this fd is registered with."
        await self.epoller.epfd.epoll_ctl(
            EPOLL_CTL.MOD, self.fd, await self.epoller.ram.ptr(EpollEvent(self.number, events)))

    async def delete(self) -> None:
        "Delete this fd from the epollfd."
        if not self.in_epollfd:
            raise Exception("already deleted from epollfd")
        await self.epoller.epfd.epoll_ctl(EPOLL_CTL.DEL, self.fd)
        self.epoller.epoll_waiter.remove_number_after_next_epoll_wait(self.number)
        self.in_epollfd = False

    async def _run(self) -> None:
        waiters: t.Dict[EPOLL, t.List[t.Tuple[EPOLL, Continuation[None]]]] = {flag: [] for flag in EPOLL}
        while True:
            try:
                ev = await self.epoller.epoll_waiter.queue.request(self.number)
            except Exception as e:
                final_exn = e
                break
            self.total_events.update(ev)
            self.status.posedge(ev)
            for val, coro in self.queue.fetch_any():
                for flag in val:
                    waiters[flag].append((val, coro))
            for flag in ev:
                to_resume = waiters[flag]
                for val, cb in list(to_resume):
                    # remove the cb from other lists it's on; cool scope there, Python
                    for waiting_flag in val:
                        waiters[waiting_flag].remove((val, cb))
                    cb.send(None)
        self.queue.close(final_exn)

    async def wait_for(self, flags: EPOLL) -> None:
        "Call epoll_wait until at least one of the passed flags is set in our status."
        if (not (self.status.mask & flags)
            and not any(self.total_events[flag] - self.consumed_events[flag] > 0
                        for flag in flags)):
            return await self.queue.request(flags)

    def consume(self, events: t.Dict[EPOLL, int]) -> None:
        """The information from these events is outdated; discard them

        Each syscall gives us information about some kinds of events, and information from
        events from epoll_wait calls before that syscall are obsoleted by this more
        up-to-date information.

        So, we call `get_current_events` before each syscall, for the set of flags that
        that syscall will inform us about. Then, after the syscall, we call `consume`, to
        indicate that the information from those events is now obsolete.

        We can't consume the events until after the syscall is done, because we want to
        allow others to optimistically perform syscalls based on these epoll events; not
        consuming the events immediately also makes us tolerate being canceled without
        performing the syscall.

        Only the events received before the syscall are known to be obsolete; events
        received during or after this syscall need to be preserved until more syscalls are
        made, to prevent deadlocks.

        The reason we need this somewhat complicated edifice is because our calls to
        epoll_wait are not synchronized with our system calls on individual file
        descriptors; calls to epoll_wait and fd system calls can happen in different
        threads and can be arbitrarily reordered relative to each other.

        Note, however, that system calls made from an individual `AsyncFileDescriptor`
        instance *are* synchronized relative to each other; those system calls are made in
        a single thread, so the system calls return in order thanks to `dneio`. This
        allows us to safely immediately call `FDStatus.negedge`/`FDStatus.posedge` after a
        system call.

        TODO if we have multiple concurrently-used `AsyncFileDescriptor` instances,
        though, they aren't synchronized, so we need to be a little more careful...

        """
        self.consumed_events.update(((flag, max(self.consumed_events[flag], count))
                                     for flag, count in events.items()))

    def get_current_events(self, flags: EPOLL) -> t.Dict[EPOLL, int]:
        "Get the current set of events from epoll_wait for these flags, for use with consume."
        return {flag: self.total_events[flag] for flag in flags}

@dataclass
class FDStatus:
    """Tracks the IO status of a file as an EPOLL mask.

    With edge-triggered epoll, we can track whether a file is ready for a given IO in
    userspace.

    Edge-triggered epoll sends us a stream of posedges for each tracked file: Each time we get
    an EPOLLET event, it means that the file is ready for the IO operations corresponding
    to the bits set in the event. We track this by just or-ing in the event's mask into
    FDStatus's mask.

    We receive negedges by getting EAGAINs (or similar equivalents) when performing
    operations on file descriptors. Each one tells us that the file is not ready for some
    specific IO operation. We track this by translating it into the corresponding EPOLL
    flag and unsetting it in FDStatus's mask.

    By following this discipline and getting every event, we know for sure at any time
    whether a file *might* be ready for IO, or if instead it's safe to block in a call to
    epoll_wait to wait for the file to be ready.

    Dropping a negedge decreases our efficiency, but is harmless, since we'll just get an
    EAGAIN the next time. Dropping a posedge is very bad and means we will deadlock.

    epoll seems to not send us a posedge if we could already know that the file descriptor
    is readable. So, for example, if we directly read an fd and we haven't already gotten
    a posedge for it, and we get a full buffer of data, then we won't get a posedge for it
    from epoll. This would cause various problems, including deadlocks if multiple threads
    were trying to read from the fd and one was blocked in epoll_wait, waiting for the
    posedge.

    Therefore, if you optimistically read an fd before getting a posedge from epoll, and
    you successfully read some data, make sure to treat that as a posedge for EPOLL.IN.
    Likewise with write and EPOLL.OUT, and any other event and the appropriate EPOLL bit.

    """
    mask: EPOLL

    def posedge(self, event: EPOLL) -> None:
        self.mask |= event

    def negedge(self, event: EPOLL) -> None:
        self.mask &= ~event

class AsyncFileDescriptor:
    """A file descriptor on which IO can be performed without blocking the thread.

    Also comes with helpful methods `AsyncFileDescriptor.write_all_bytes` and
    `AsyncFileDescriptor.read_some_bytes` to abstract over memory allocation.

    We always wait for a posedge to come back from epoll before trying to read. This is
    not necessarily too pessimistic, because as soon as we have a single posedge, we will
    keep reading in a loop, as long as data keeps coming through.

    """
    @staticmethod
    async def make(epoller: Epoller, ram: RAM, fd: FileDescriptor) -> AsyncFileDescriptor:
        """Make an AsyncFileDescriptor; make sure to call this with only O.NONBLOCK file descriptors.

        It won't actually break anything if this is called with file descriptors not in
        NONBLOCK mode; it just means that they'll block when we go to read, which is
        probably not what the user wants.

        """
        epolled = await epoller.register(
            fd, EPOLL.IN|EPOLL.OUT|EPOLL.RDHUP|EPOLL.PRI|EPOLL.ERR|EPOLL.HUP|EPOLL.ET,
        )
        return AsyncFileDescriptor(ram, fd, epolled)

    def __init__(self, ram: RAM, handle: FileDescriptor,
                 epolled: EpolledFileDescriptor,
    ) -> None:
        "Don't construct directly; use the AsyncFileDescriptor.make constructor instead."
        self.ram = ram
        self.handle = handle
        "The underlying FileDescriptor for this AFD, used for all system calls"
        self.epolled = epolled

    def __str__(self) -> str:
        return f"AsyncFileDescriptor({self.epolled})"

    async def make_new_afd(self, fd: FileDescriptor) -> AsyncFileDescriptor:
        """Use the Epoller and RAM in this AsyncFD to make a new `AsyncFileDescriptor` for `fd`

        Make sure that `fd` is already in non-blocking mode;
        such as by accepting it with the `SOCK.NONBLOCK` flag.

        This doesn't steal any resources from the original AFD; it's just a convenience method,
        most useful when calling accept() and wanting to create new AFDs out of the resulting FDs.

        """
        return await AsyncFileDescriptor.make(self.epolled.epoller, self.ram, fd)

    async def wait_for_rdhup(self) -> None:
        "Call epoll_wait until this file descriptor has a hangup."
        await self.epolled.wait_for(EPOLL.RDHUP|EPOLL.HUP)

    async def read(self, ptr: Pointer) -> t.Tuple[Pointer, Pointer]:
        "Call `FileDescriptor.read` without blocking the thread."
        while True:
            await self.epolled.wait_for(EPOLL.IN|EPOLL.RDHUP|EPOLL.HUP|EPOLL.ERR)
            current_events = self.epolled.get_current_events(EPOLL.IN|EPOLL.RDHUP|EPOLL.HUP|EPOLL.ERR)
            try:
                return (await self.handle.read(ptr))
            except OSError as e:
                self.epolled.consume(current_events)
                if e.errno == errno.EAGAIN:
                    self.epolled.status.negedge(EPOLL.IN|EPOLL.RDHUP|EPOLL.HUP|EPOLL.ERR)
                else:
                    self.epolled.status.posedge(EPOLL.ERR)
                    raise
            else:
                self.epolled.consume(current_events)
                self.epolled.status.posedge(EPOLL.IN|EPOLL.RDHUP|EPOLL.HUP)

    async def read_some_bytes(self, count: int=4096) -> bytes:
        """Read at most count bytes; possibly less, if we have a partial read.

        This allocates on each call. For some applications, you may want to avoid the cost of
        allocation, by instead allocating a buffer with `Thread.malloc` up front and reusing it
        across multiple calls to `AsyncFileDescriptor.read`.

        """
        ptr = await self.ram.malloc(bytes, count)
        valid, _ = await self.read(ptr)
        return await valid.read()

    async def write(self, buf: Pointer) -> t.Tuple[Pointer, Pointer]:
        """Call `FileDescriptor.write` without blocking the thread.

        Note that this doesn't retry partial writes, which are always a possibility, so you should
        make sure to do that yourself, or use `AsyncFileDescriptor.write_all`.

        """
        while True:
            await self.epolled.wait_for(EPOLL.OUT|EPOLL.ERR)
            current_events = self.epolled.get_current_events(EPOLL.OUT|EPOLL.ERR)
            try:
                return await self.handle.write(buf)
            except OSError as e:
                self.epolled.consume(current_events)
                if e.errno == errno.EAGAIN:
                    self.epolled.status.negedge(EPOLL.OUT|EPOLL.ERR)
                else:
                    self.epolled.status.posedge(EPOLL.ERR)
                    raise
            else:
                self.epolled.consume(current_events)
                self.epolled.status.posedge(EPOLL.OUT)

    async def write_all(self, to_write: Pointer) -> None:
        """Write all of this pointer to the fd, retrying on partial writes until complete.

        You might want to not use this, if you want to react to a partial write in some special way.
        For example, `rsyscall.memory.socket_transport.SocketMemoryTransport` starts a `recv`
        immediately after a partial write, before retrying the write, for increased parallelism.

        """
        while to_write.size() > 0:
            written, to_write = await self.write(to_write)

    async def write_all_bytes(self, buf: bytes) -> None:
        """Write all these bytes to the fd, retrying on partial writes until complete.

        This allocates and performs a store to memory on each call. This is inefficient if you
        already have an initialized pointer for the value, or if you already have an allocated
        buffer that you can use to store the value, which you can also reuse for other values.
        In those cases, you might want to use `AsyncFileDescriptor.write_all`.

        """
        ptr = await self.ram.ptr(buf)
        await self.write_all(ptr)

    @t.overload
    async def accept(self, flags: SOCK=SOCK.NONE) -> FileDescriptor: ...
    @t.overload
    async def accept(self, flags: SOCK, addr: WrittenPointer[Sockbuf[T_sockaddr]]
    ) -> t.Tuple[FileDescriptor, WrittenPointer[Sockbuf[T_sockaddr]]]: ...

    async def accept(self, flags: SOCK=SOCK.NONE, addr: t.Optional[WrittenPointer[Sockbuf[T_sockaddr]]]=None
    ) -> t.Union[FileDescriptor, t.Tuple[FileDescriptor, WrittenPointer[Sockbuf[T_sockaddr]]]]:
        "Call accept without blocking the thread."
        while True:
            await self.epolled.wait_for(EPOLL.IN|EPOLL.HUP|EPOLL.ERR)
            current_events = self.epolled.get_current_events(EPOLL.IN|EPOLL.HUP|EPOLL.ERR)
            try:
                if addr is None:
                    return (await self.handle.accept(flags))
                else:
                    return (await self.handle.accept(flags, addr))
            except OSError as e:
                self.epolled.consume(current_events)
                if e.errno == errno.EAGAIN:
                    self.epolled.status.negedge(EPOLL.IN|EPOLL.HUP|EPOLL.ERR)
                else:
                    self.epolled.status.posedge(EPOLL.HUP|EPOLL.ERR)
                    raise
            else:
                self.epolled.consume(current_events)
                self.epolled.status.posedge(EPOLL.IN)

    async def accept_addr(self, flags: SOCK=SOCK.NONE) -> t.Tuple[FileDescriptor, Sockaddr]:
        "Call accept with a buffer for the address, and return the resulting fd and address."
        written_sockbuf = await self.ram.ptr(Sockbuf(await self.ram.malloc(SockaddrStorage)))
        fd, sockbuf = await self.accept(flags, written_sockbuf)
        addr = (await (await sockbuf.read()).buf.read()).parse()
        return fd, addr

    async def bind(self, addr: WrittenPointer[Sockaddr]) -> None:
        "Call bind; bind already doesn't block the thread."
        await self.handle.bind(addr)

    async def connect(self, addr: WrittenPointer[Sockaddr]) -> None:
        "Call connect without blocking the thread."
        try:
            # Note that an unconnected socket, at least with AF.INET SOCK.STREAM,
            # will have EPOLL.OUT|EPOLL.HUP set when added to epoll, before calling connect.
            current_events = self.epolled.get_current_events(EPOLL.OUT)
            await self.handle.connect(addr)
        except OSError as e:
            self.epolled.consume(current_events)
            self.epolled.status.negedge(EPOLL.OUT)
            if e.errno == errno.EINPROGRESS:
                await self.epolled.wait_for(EPOLL.OUT)
                current_events = self.epolled.get_current_events(EPOLL.OUT)
                sockbuf = await self.ram.ptr(Sockbuf(await self.ram.malloc(Int32)))
                retbuf = await self.handle.getsockopt(SOL.SOCKET, SO.ERROR, sockbuf)
                err = await (await retbuf.read()).buf.read()
                self.epolled.consume(current_events)
                if err == 0:
                    self.epolled.status.posedge(EPOLL.OUT)
                else:
                    self.epolled.status.posedge(EPOLL.ERR)
                    try:
                        raise OSError(err, os.strerror(err))
                    except OSError as exn:
                        exn.filename = self.handle
                        if hasattr(addr, 'value'):
                            exn.filename2 = addr.value
                        raise
            else:
                raise
        else:
            self.epolled.consume(current_events)
            self.epolled.status.posedge(EPOLL.OUT)

    def with_handle(self, fd: FileDescriptor) -> AsyncFileDescriptor:
        """Return a new AFD using this new FD handle for making syscalls.

        This is useful when we want to change what task we're making syscalls in when
        using this AFD.

        """
        return AsyncFileDescriptor(self.ram, fd, self.epolled)

    async def close(self) -> None:
        "Remove this FD from Epoll and invalidate the FD handle."
        await self.epolled.delete()
        await self.handle.invalidate()

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args, **kwargs) -> None:
        await self.close()


################################################################################
# Miscellaneous helpers

class AsyncReadBuffer:
    """A buffer for parsing variable-length streaming data.

    When reading data from a stream such as a pipe or TCP connection, data is not
    delivered to us from the kernel in nicely-separated records. We need to rebuffer the
    data so that it can be parsed. That's what this class does; and it provides a few
    helper methods to make it easier to read and parse such streams.

    """
    def __init__(self, fd: AsyncFileDescriptor) -> None:
        self.fd = fd
        self.buf = b""
        self.unread_ptr: t.Optional[Pointer] = None

    async def _read(self) -> bytes:
        "Read some bytes; return None on EOF."
        if self.unread_ptr is None:
            ptr = await self.fd.ram.malloc(bytes, 4096)
            self.unread_ptr, _ = await self.fd.read(ptr)
        if self.unread_ptr.size():
            data = await self.unread_ptr.read()
        else:
            data = b''
        self.unread_ptr = None
        if len(data) == 0:
            raise EOFError
        else:
            return data

    async def read_length(self, length: int) -> bytes:
        "Read exactly this many bytes; raises on EOF."
        while len(self.buf) < length:
            data = await self._read()
            self.buf += data
        section = self.buf[:length]
        self.buf = self.buf[length:]
        return section

    async def read_cffi(self, name: str) -> t.Any:
        "Read, parse, and return this fixed-size cffi type."
        size = ffi.sizeof(name)
        try:
            data = await self.read_length(size)
        except EOFError as e:
            e.args = ("got EOF while expecting to read a", name, "of size", size)
            raise
        nameptr = name + '*'
        dest = ffi.new(nameptr)
        # ffi.cast drops the reference to the backing buffer, so we have to copy it
        src = ffi.cast(nameptr, ffi.from_buffer(data))
        ffi.memmove(dest, src, size)
        return dest[0]

    async def read_struct(self, cls: t.Type[T_fixed_size]) -> T_fixed_size:
        "Read one fixed-size struct from the buffer, or return None if that's not possible"
        size = cls.sizeof()
        try:
            data = await self.read_length(size)
        except EOFError as e:
            e.args = ("got EOF while expecting to read a", cls, "of size", size)
            raise
        return cls.get_serializer(self.fd.handle.task).from_bytes(data)

    async def read_length_prefixed_string(self) -> bytes:
        "Read a bytestring which is prefixed with a 64-bit native-byte-order size."
        elem_size = await self.read_cffi('size_t')
        try:
            elem = await self.read_length(elem_size)
        except EOFError as e:
            e.args = ("got EOF while expecting to read environment element of length", elem_size)
        return elem

    async def read_length_prefixed_array(self, length: int) -> t.List[bytes]:
        "Read an array, prefixed with its size, of bytestrings, each prefixed with their size."
        ret: t.List[bytes] = []
        for _ in range(length):
            ret.append(await self.read_length_prefixed_string())
        return ret

    async def read_envp(self, length: int) -> t.Dict[str, str]:
        """Read a size-prefixed array of size-prefixed bytestrings, with each bytestring containing '=', into a dict.

        This is the format we expect for envp, which is written to us on startup by
        non-child threads.

        We assume that the environment is correctly formed; that is, each element contains
        '='.  If that's not true, we'll throw an exception.  But it would be quite unusual
        to have an incorrectly formed environment variable, so it's not too concerning.

        """
        raw = await self.read_length_prefixed_array(length)
        environ: t.Dict[str, str] = {}
        for elem in raw:
            key, val = elem.split(b"=", 1)
            environ[os.fsdecode(key)] = os.fsdecode(val)
        return environ

    async def read_until_delimiter(self, delim: bytes) -> t.Optional[bytes]:
        "Read and return all bytes until the specified delimiter, stripping the delimiter; on EOF, return None."
        while True:
            try:
                i = self.buf.index(delim)
            except ValueError:
                pass
            else:
                section = self.buf[:i]
                # skip the delimiter
                self.buf = self.buf[i+1:]
                return section
            # buf contains no copies of "delim", gotta read some more data
            data = await self._read()
            if data is None:
                return None
            self.buf += data

    async def read_line(self) -> bytes:
        "Read and return a line, stripping the newline character."
        ret = await self.read_until_delimiter(b"\n")
        if ret is None:
            raise EOFError("hangup before reading full line")
        return ret

    async def read_netstring(self) -> bytes:
        "Read a netstring, as defined by DJB."
        length_bytes = await self.read_until_delimiter(b':')
        if length_bytes is None:
            raise EOFError("hangup before reaching colon at end of netstring size")
        length = int(length_bytes)
        try:
            data = await self.read_length(length)
        except EOFError as e:
            e.args = ("hangup before we read netstring data of size", length)
            raise
        try:
            comma = await self.read_length(1)
        except EOFError as e:
            e.args = ("hangup before comma at end of netstring",)
            raise
        if comma != b",":
            raise Exception("bad netstring delimiter", comma)
        return data



"""Like the stdlib codeop module, but returning an AST instead.

This is useful because we can properly deal with `await`s at the AST level.

We lack functionality precisely equivalent to `codeop.Compile` or `codeop.CommandCompiler`,
because the AST object returned from `compile(ONLY_AST)` doesn't expose the information to
us about what `__future__` statements the compile process has seen. To properly implement
those classes, either the return value of `compile(ONLY_AST)` needs to contain that
information, or we need to reimplement the simple `__future__` statement scanner contained
in the Python core.

"""
import codeop
import ast
import typing as t

def _ast_compile(source, filename, symbol) -> t.Any:
    PyCF_DONT_IMPLY_DEDENT = codeop.PyCF_DONT_IMPLY_DEDENT # type: ignore
    return compile(source, filename, symbol, ast.PyCF_ONLY_AST|PyCF_DONT_IMPLY_DEDENT)

def ast_compile_command(source: str, filename="<input>", symbol="single") -> t.Any:
    "Like codeop.compile_command, but returns an AST instead."
    _maybe_compile = codeop._maybe_compile # type: ignore
    return _maybe_compile(_ast_compile, source, filename, symbol)

def ast_compile_interactive(source: str) -> t.Optional[ast.Interactive]:
    "Compiles this single interactive statement into an AST"
    return ast_compile_command(source, "<input>", "single")



import dns.rdata
import dns.tokenizer

class LUA(dns.rdata.Rdata):
    __slots__ = ['type', 'lua']

    def __init__(self, rdclass, rdtype, type: str, lua: str) -> None:
        super().__init__(rdclass, rdtype)
        self.type = type
        self.lua = lua

    def to_text(self, origin=None, relativize=True, **kw):
        return self.type + " \"" + self.lua + "\""

    @classmethod
    def from_text(cls, rdclass, rdtype, tok: dns.tokenizer.Tokenizer, origin=None, relativize=True):
        typ = tok.get_identifier()
        # now we spin in a loop, getting tokens, skipping whitespace...
        # until we get an eol?
        # oh, until we see an eol token, yes
        cur = tok.get()
        if cur.is_eol_or_eof():
            raise Exception("expecting some lua code")
        lua = cur.value
        while True:
            cur = tok.get()
            if cur.is_eol_or_eof():
                break
            lua += " " + cur.value
        return cls(rdclass, rdtype, typ, lua)



"""A typed API for generically prompting the user for a typed value, typically by dropping to a REPL

The metaphor we use is "wishing" for a value of some type, and then getting back a value
of that type. We use the wish function and the Wish class to do this. We follow many
patterns of exception handling, so `wish` can be thought of as a version of `raise` which
returns a value, and `Wish` can be thought of as an Exception object; indeed, it inherits
from Exception. Like `raise` and exception handlers, `wish` will always invoke the closest
bound wish handler.

The `wish` API is sufficiently general to cover a variety of different uses. Most
basically, it can be used to drop to a REPL at any time so that the user may interact with
the program. Since `wish` is async, this can be used even in running applications without
disturbing the rest of the application, unlike alternatives like `breakpoint`.

Another application is repairing broken systems. We can model "repair" of a broken system
as wishing for a fixed version of a broken value, which will be interactively corrected by
the user. If we make such a wish, we might want the code calling wish to perform
additional automated checks on the returned value, and wish again if the checks fail.

Yet another application is simple prompting of the user for values we cannot produce
programmatically, or prompting for the user to take actions we cannot perform
programmatically. We might prompt the user for some difficult-to-encode information about
the state of the world. Or we might prompt the user to perform some action manually that
is difficult for us to do automatically. Because `wish` invokes a wish handler, we can
automate these actions after the fact, and modify our automation to our needs, without
changing the module that calls `wish`.

To wish for a value, we create a value of class Wish, which should contain the type object
that we're wishing for, and pass it to the wish function. wish will capture a traceback of
our stack, attach it to the Wish, present that Wish to the user, and return the value
returned by the user.

When we wish for a value, we invoke the closest bound WishGranter; that handles prompting
the user and providing some UI for the user to produce the value. The WishGranter may
choose to typecheck the value returned by the user to ensure that it matches the value
requested in the Wish.  If the WishGranter provides a REPL, it may want to extract local
variables from the traceback captured by wish, and present them to the user. If no
WishGranter is bound, we simply raise the Wish as an exception.

This library does not provide any WishGranters; there are several in the rsyscall.wish module.
That module provides only console REPL WishGranters, but the API is entirely extensible,
so other ways to present a REPL, or even entirely different UIs, are possible.
For example, a REPL might be presented with a web UI; perhaps even a Jupyter notebook.

A WishGranter might also be entirely or partially automatic; it could respond to different
types of Wishes automatically, based on the type wished for, the subclass of Wish used,
the message, or other attributes, and propagate other wishes up to the WishGranter bound
above it.

"""
from __future__ import annotations
from contextvars import ContextVar
import abc
import inspect
import sys
import traceback
import types
import typing as t

__all__ = [
    'Wish',
    'WishGranter',
    'wish',
    'my_wish_granter',
]

# TODO should we inherit from BaseException or Exception?
T = t.TypeVar('T')
class Wish(BaseException, t.Generic[T]):
    """A request for a value of type `return_type`

    We inherit from BaseException. When wish is called on this value, it will fill in the
    exception fields on this value, in the same way that raise would fill in the exception
    fields of any exception.

    A user can further inherit from this class to add more information to the Wish, in the
    same way one would inherit from Exception to create a more specific Exception.

    """
    return_type: t.Type[T]

    def __init__(self, return_type: t.Type[T], *args) -> None:
        self.return_type = return_type
        super().__init__(*args)

my_wish_granter: ContextVar[WishGranter] = ContextVar('my_wish_granter')
"""The WishGranter currently active for our scope.

To use a new WishGranter in some scope, set my_wish_granter to a new value, then reset it
back to the old value once you leave that scope.  A contextmanager would likely be helpful
for this.

"""

class WishGranter:
    """An object capable of responding to a wish.

    A WishGranter is not active until it is bound. When wish is called, the closest bound
    WishGranter is used to satisfy the wish. The WishGranter may itself choose to call
    wish and propagate the wish up the stack. For example, a partially automated
    WishGranter might be able to handle only certain classes of Wish, and fall back to
    wishing again when it sees other classes.

    WishGranters are bound to my_wish_granter to become active. This merely means they
    will be used by `wish`; no actual change occurs to the WishGranter object.

    Since my_wish_granter is a ContextVar, it's useful to make an analogy to exception
    handlers, because the resolution process for ContextVars, and therefore WishGranters,
    works the same way.

    """
    @abc.abstractmethod
    async def wish(self, wish: Wish[T]) -> T:
        "Satisfy this wish, returning a value of the type requested by the wish"
        pass

def _frames_to_traceback(frames: t.List[types.FrameType]) -> t.Optional[types.TracebackType]:
    "Translate a list of frames (which can be obtained from the inspect module) to a traceback"
    tb = None
    for frame in frames:
        tb = types.TracebackType(tb, frame, frame.f_lasti, frame.f_lineno)
    return tb

# TODO should switch bool to typing_extensions.Literal[False]
async def wish(wish: Wish[T], *, from_exn: t.Union[BaseException, None, bool]=False) -> T:
    """Wish for some value, as specified by the passed Wish, and get that value

    Pass `from_exn=exn` to explicitly set the cause/context for the Wish to the exception
    `exn`; this is identical in behavior to doing `raise wish from exn` with exceptions.
    Likewise, you can pass `from_exn=None` to suppress the exception cause/context.

    We use the WishGranter currently bound to my_wish_granter in our ContextVar context to
    perform the wish. This is directly analogous to using the closest bound exception
    handler to handle an exception, and works the same way.

    """
    if not isinstance(wish, Wish):
        raise Exception("wishes should be of type Wish, not", wish)
    raising_exception = sys.exc_info()[1]
    if not isinstance(from_exn, bool):
        wish.__cause__ = from_exn
    elif raising_exception:
        wish.__context__ = raising_exception

    wish.__traceback__ = _frames_to_traceback([record.frame for record in inspect.stack()[1:]])

    wish_granter = my_wish_granter.get()
    ret = await wish_granter.wish(wish)
    return ret



"""Monitoring child processes in a non-blocking manner

The API for monitoring child processes is one of the worst-designed parts of
Unix, and Linux inherits that bad design. Still, we've tried to provide a usable
interface here.

Our basic approach is to use a signalfd waiting for SIGCHLD to learn when a
state change may have happened, then call waitid until we run out of state
changes. This is a common way to handle it, but we have a few changes.

Contrary to the obvious implementation, we don't use waitid(P.ALL).  Instead, we
make an individual waitid(P.PID) call for each process we're currently
monitoring for state changes.

P.ALL is unrecoverably broken and racy in many situations. It's fundamentally
the wrong API: You can't wait for only a subset of processes, you always have to
handle every single state change that could happen for any process. And worse:
When you handle the death state change for a child, that child pid is freed and
can be reused; so you must synchronize against anything that wants to operate on
child pids, to prevent pid wrap races.

P.ALL is especially bad when considering the possibility of new children
appearing.  If we call clone in two threads while a third calls waitid, and one
of the created child processes dies immediately, there is no way to prevent a
race where waitid gets the death state change before the second clone completes,
the pid wraps, the same pid is returned for the second clone, and we don't know
which child process is alive and which is dead. Similar races can happen if
we're pid 1 or a subreaper and we call clone.

Issuing a batch of P.PID calls instead is much better: We can wait for a subset
of pids, and so all these worries about races go away.  We still need to
synchronize between usage of the pid and calls to waitid, but that
synchronization happens for the individual child process, not globally across
all our child processes. And the implementation becomes much simpler: Each
object monitoring an individual child process just calls waitid(P.PID) and
checks the result, instead of having to wait on some central coordinator to call
waitid(P.ALL) and send back the state changes.

We also employ another trick: We use CLONE.PARENT to centralize child
monitoring.  When thread A creates child thread B, we want thread B to be able
to later create children of its own, e.g. process C.  Normally, we would create
a new signalfd in thread B so that it can monitor its children. Instead, when
thread B clones a new child process C, we use CLONE.PARENT so that thread A
becomes the parent of new child process C, then we monitor process C from thread
A instead of thread B.  This lowers the original creation cost of thread B
(since we don't have to make a new AsyncSignalfd, which would necessitate a new
epoller - see the docstring for AsyncSignalfd) and also improves the efficiency
of child monitoring through centralization into thread A.

"""
from __future__ import annotations
from dataclasses import dataclass
from dneio import RequestQueue, reset, Event
from rsyscall.epoller import Epoller, AsyncFileDescriptor
from rsyscall.handle import WrittenPointer, Pointer, Stack, FutexNode, Task, Pointer, ChildProcess
from rsyscall.memory.ram import RAM
from rsyscall.near.sysif import SyscallError
from rsyscall.sched import CLONE
from rsyscall.signal import SIG, Sigset, Siginfo
import trio
import contextlib
import typing as t
import logging
logger = logging.getLogger(__name__)

from rsyscall.signal import SignalBlock
from rsyscall.sys.signalfd import SFD, SignalfdSiginfo
from rsyscall.sys.wait import CLD, ChildState, W, CalledProcessError

class AsyncSignalfd:
    """A signalfd, registered on epoll, with a SignalBlock for the appropriate signals

    Note that signalfd and epoll have some quirks when used
    together. Specifically, a signalfd registered on an epollfd in one
    process will cause severe breakage, including deadlocks, when
    calling epoll_wait on that epollfd in another process.

    For us, this means that if we're using a signalfd for some
    process, we must also use an epoller that will be waited-on in
    that same process. We take care at call sites of
    AsyncSignalfd.make so that this is true.

    It would be very nice if this bug in signalfd was fixed, that
    would allow a single central process to monitor for signals in
    many other processes.

    """
    @classmethod
    async def make(cls, ram: RAM, task: Task, epoller: Epoller, mask: Sigset,
                   *, signal_block: SignalBlock=None,
    ) -> AsyncSignalfd:
        """Make a signalfd and register it on the epoller, possibly blocking signals

        If the signals are already blocked, the user can pass in a SignalBlock to
        represent that, and save the need to make the SignalBlock.

        """
        if task is not epoller.epoll_waiter.epfd.task:
            raise Exception("signalfd task and epoll_waiter task must be the same")
        if signal_block is None:
            async def op(sem: RAM) -> t.Tuple[WrittenPointer[Sigset], Pointer[Sigset]]:
                return await sem.ptr(mask), await sem.malloc(Sigset)
            sigset_ptr, oldset_ptr = await ram.perform_batch(op)
            signal_block = await task.sigmask_block(sigset_ptr, oldset_ptr)
            await task.read_oldset_and_check()
        else:
            sigset_ptr = signal_block.newset
        afd = await AsyncFileDescriptor.make(epoller, ram, await task.signalfd(sigset_ptr, SFD.NONBLOCK))
        return cls(afd, signal_block, await ram.malloc(SignalfdSiginfo))

    def __init__(self,
                 afd: AsyncFileDescriptor,
                 signal_block: SignalBlock,
                 buf: Pointer[SignalfdSiginfo],
    ) -> None:
        "Use the constructor method AsyncSignalfd.make"
        self.afd = afd
        self.signal_block = signal_block
        self.buf = buf
        self.next_signal = Event()
        reset(self._run())

    async def _run(self) -> None:
        while True:
            try:
                valid, rest = await self.afd.read(self.buf)
            except SyscallError as syscall_error:
                final_exn = syscall_error
                break
            ev, self.next_signal = self.next_signal, Event()
            # we discard the signal information...
            ev.set()
            self.buf = valid + rest
        self.next_signal.close(final_exn)

class AsyncChildProcess:
    "A child process which can be monitored without blocking the thread"
    def __init__(self, process: ChildProcess, ram: RAM, sigchld_sigfd: AsyncSignalfd) -> None:
        self.process = process
        self.ram = ram
        self.sigchld_sigfd = sigchld_sigfd
        self.next_sigchld: t.Optional[Event] = None

    def __repr__(self) -> str:
        name = type(self).__name__
        return f'{name}({self.process})'

    async def _waitid_nohang(self) -> t.Optional[ChildState]:
        if self.process.unread_siginfo:
            # if we performed a waitid before, and it contains an event, we don't need to
            # waitid again.
            result = await self.process.read_siginfo()
            # but if there's no event in this previous waitid, we need to waitid now; if
            # we don't, we might erroneously block waiting for a SIGCHLD that happened
            # between the previous waitid and now, and was consumed at that time.
            if result:
                return result
        siginfo_buf = await self.ram.malloc(Siginfo)
        await self.process.waitid(W.EXITED|W.STOPPED|W.CONTINUED|W.NOHANG, siginfo_buf)
        return await self.process.read_siginfo()

    async def waitpid(self, options: W) -> ChildState:
        "Wait for a child state change in this child, like waitid(P.PID)"
        if options & W.EXITED and self.process.death_state:
            # TODO this is not really the actual behavior of waitpid...
            # if the child is already dead we'd get an ECHLD not the death state change again.
            return self.process.death_state
        while True:
            # If a previous call has given us a next_sigchld to wait on, then wait we shall.
            if self.next_sigchld:
                await self.next_sigchld.wait()
                # we shouldn't wait for SIGCHLD the next time we're called, we should eagerly call
                # waitid, since there may still be state changes to fetch.
                self.next_sigchld = None
            # We have to save this signal event before calling waitid, otherwise we may deadlock: If
            # a SIGCHLD is delivered while we're calling waitid, then saved_sigchld will be
            # different from self.sigchld_sigfd.next_signal after the waitid; and if we use the
            # value of self.sigchld_sigfd.next_signal after the waitid, we'll be waiting for a
            # SIGCHLD that will never come.
            saved_sigchld = self.sigchld_sigfd.next_signal
            state_change = await self._waitid_nohang()
            if state_change is not None:
                if state_change.state(options):
                    return state_change
                else:
                    # TODO we shouldn't discard the state change here if we're not waiting for it;
                    # unfortunately doing it right will require a lot of refactoring of waitid
                    pass
            else:
                # we know for sure that there will only be state changes fetchable by waitid after
                # waiting for this SIGCHLD event. note that this event may have already happened, if
                # we received a SIGCHLD while calling waitid.
                self.next_sigchld = saved_sigchld

    async def wait(self, options: W=W.EXITED|W.STOPPED|W.CONTINUED) -> ChildState:
        return await self.waitpid(options)

    async def check(self) -> ChildState:
        "Wait for this child to die, and once it does, throw `rsyscall.sys.wait.CalledProcessError` if it didn't exit cleanly"
        try:
            death = await self.waitpid(W.EXITED)
        except trio.Cancelled:
            await self.kill(SIG.TERM)
            raise
        if not death.clean():
            if self.process.command:
                raise CalledProcessError(death, self.process.command)
            else:
                raise CalledProcessError(death)
            pass
        return death

    async def kill(self, sig: SIG=SIG.KILL) -> None:
        "Send a signal to this child"
        if self.process.unread_siginfo:
            await self.process.read_siginfo()
        await self.process.kill(sig)

    async def killpg(self, sig: SIG=SIG.KILL) -> None:
        "Send a signal to the process group corresponding to this child"
        if self.process.unread_siginfo:
            await self.process.read_siginfo()
        await self.process.killpg(sig)

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args, **kwargs) -> None:
        if self.process.death_state:
            pass
        else:
            await self.kill()
            await self.waitpid(W.EXITED)

@dataclass
class ChildProcessMonitor:
    """Contains all that is needed to create an AsyncChildProcess from a ChildProcess

    We also know what arguments to pass to clone so that an AsyncChildProcess may be created.

    Use ChildProcessMonitor.make to create.

    """
    sigfd: AsyncSignalfd
    ram: RAM
    cloning_task: Task
    use_clone_parent: bool

    @staticmethod
    async def make(ram: RAM, task: Task, epoller: Epoller,
                   *, signal_block: SignalBlock=None,
    ) -> ChildProcessMonitor:
        """Make a ChildProcessMonitor, possibly blocking signals

        If the signals are already blocked, the user can pass in a SignalBlock to
        represent that, and save the need to make the SignalBlock.

        """
        sigfd = await AsyncSignalfd.make(ram, task, epoller, Sigset({SIG.CHLD}), signal_block=signal_block)
        return ChildProcessMonitor(sigfd, ram, task, use_clone_parent=False)

    def inherit_to_child(self, child_task: Task) -> ChildProcessMonitor:
        """Create a new instance that will clone children from the passed-in task

        This requires, and checks, that the passed-in task is a child of the process which
        is used for monitoring in the current instance. Then the functionality is
        implemented by just using CLONE.PARENT when calling clone in child_task, and
        thereby delegating responsibility for monitoring to the parent.

        """
        if child_task.parent_task is not self.sigfd.afd.handle.task:
            raise Exception("task", child_task, "with parent_task", child_task.parent_task,
                            "is not our child; we're", self.sigfd.afd.handle.task)
        # 1. We know that child_task is a child process of self.sigfd.afd.handle.task.
        # 2. That means if we use CLONE_PARENT in child_task, the resulting processes will also be
        # child processes of self.sigfd.afd.handle.task.
        # 3. Therefore self.sigfd will be notified if and when those future child processes have some state change.
        # 4. Therefore we can use self.sigfd to create AsyncChildProcesses for those future child processes.
        return ChildProcessMonitor(self.sigfd, self.ram, child_task, use_clone_parent=True)

    def add_child_process(self, process: ChildProcess) -> AsyncChildProcess:
        """Create an AsyncChildProcess which monitors the passed-in ChildProcess.

        We check that the passed-in process is in fact a child of the task associated with
        this ChildProcessMonitor before returning the AsyncChildProcess; otherwise the
        AsyncChildProcess wouldn't be woken up at the correct times to read child status
        changes.

        """
        if process.task is not self.sigfd.afd.handle.task:
            raise Exception("process", process, "with parent task", process.task,
                            "is not our child; we're", self.sigfd.afd.handle.task)
        proc = AsyncChildProcess(process, self.ram, self.sigfd)
        return proc

    async def clone(self, flags: CLONE,
                    child_stack: t.Tuple[Pointer[Stack], WrittenPointer[Stack]],
                    ctid: t.Optional[Pointer[FutexNode]]=None) -> AsyncChildProcess:
        """Call `clone` with these arguments and return an AsyncChildProcess monitoring the resulting child

        We'll use CLONE.PARENT if necessary to create a ChildProcess that is monitorable
        by this ChildProcessMonitor.

        """
        if self.use_clone_parent:
            flags |= CLONE.PARENT
        process = await self.cloning_task.clone(flags|SIG.CHLD, child_stack, None, ctid, None)
        return self.add_child_process(process)



from dneio import RequestQueue, reset, Event
from rsyscall.tests.trio_test_case import TrioTestCase
import outcome
import trio

class MyException(Exception):
    pass

async def sleep_and_throw() -> None:
    async with trio.open_nursery() as nursery:
        async def thing1() -> None:
            await trio.sleep(0)
            raise MyException("ha ha")
        async def thing2() -> None:
            await trio.sleep(1000)
        nursery.start_soon(thing1)
        nursery.start_soon(thing2)

class TestConcurrency(TrioTestCase):
    async def test_nursery(self) -> None:
        async with trio.open_nursery() as nursery:
            async def a1() -> None:
                await trio.sleep(10)
            async def a2() -> None:
                try:
                    await sleep_and_throw()
                except MyException:
                    pass
                finally:
                    nursery.cancel_scope.cancel()
            nursery.start_soon(a1)
            nursery.start_soon(a2)

    async def test_nest_cancel_inside_shield(self) -> None:
        "If we cancel_scope.cancel() inside a CancelScope which is shielded, it works."
        with trio.CancelScope(shield=True):
            async with trio.open_nursery() as nursery:
                nursery.start_soon(trio.sleep_forever)
                nursery.cancel_scope.cancel()

def failing_function(x):
    raise Exception("failed with", x)

class TestConcur(TrioTestCase):
    async def _first_runner(self, queue: RequestQueue) -> None:
        while True:
            many = await queue.get_many()
            for val, coro in many[::-1]:
                if val == 1337:
                    try:
                        failing_function(x) # type: ignore
                    except Exception as e:
                        coro.resume(outcome.Error(e))
                else:
                    coro.resume(outcome.Value(val + 10))

    async def _second_runner(self, queue: RequestQueue[int, int]) -> None:
        while True:
            many = await queue.get_many()
            for val, coro in many[::-1]:
                try:
                    result = await self.queue.request(val + 100)
                except Exception as e:
                    coro.resume(outcome.Error(e))
                else:
                    coro.resume(outcome.Value(result))

    async def asyncSetUp(self) -> None:
        self.queue = RequestQueue[int, int]()
        reset(self._first_runner(self.queue))
        self.second_queue = RequestQueue[int, int]()
        reset(self._second_runner(self.second_queue))

    async def test_parallel(self) -> None:
        async def req(i: int):
            ret = await self.queue.request(i)
            self.assertEqual(i+10, ret)
        async with trio.open_nursery() as nursery:
            for i in range(3):
                nursery.start_soon(req, i)

    async def test_through_multiple(self) -> None:
        self.assertEqual(await self.second_queue.request(1), 111)

    async def test_event(self) -> None:
        ev = Event()
        async with trio.open_nursery() as nursery:
            @nursery.start_soon
            async def foo():
                await trio.sleep(0)
                ev.set()
            await ev.wait()

    async def test_failure(self) -> None:
        should_fail = 1237
        def length_traceback(tb) -> int:
            if tb is None:
                return 0
            else:
                return 1 + length_traceback(tb.tb_next)
        try:
            await self.second_queue.request(should_fail)
        except Exception as e:
            self.assertLess(length_traceback(e.__traceback__), 10)




from __future__ import annotations

from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall.tasks.stub import *

import rsyscall.nix as nix
from rsyscall import local_thread

from rsyscall.tests.utils import do_async_things
from rsyscall.command import Command
from rsyscall.stdlib import mkdtemp

import os

class TestStub(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.tmpdir = await mkdtemp(self.local, "test_stub")
        # make sure that this name doesn't collide with shell builtins
        # so it can be run from the shell in test_read_stdin
        self.stub_name = "dummy_stub"
        self.server = await StubServer.make(self.local, self.tmpdir, self.stub_name)
        self.thread = await self.local.clone()

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_exit(self) -> None:
        command = Command(self.tmpdir/self.stub_name, [self.stub_name], {})
        child = await self.thread.exec(command)
        self.nursery.start_soon(child.check)
        argv, new_thread = await self.server.accept()
        await new_thread.exit(0)

    async def test_async(self) -> None:
        command = Command(self.tmpdir/self.stub_name, [self.stub_name], {})
        child = await self.thread.exec(command)
        self.nursery.start_soon(child.check)
        argv, new_thread = await self.server.accept()
        await do_async_things(self, new_thread.epoller, new_thread)

    async def test_read_stdin(self) -> None:
        data_in = "hello"
        command = self.thread.environ.sh.args("-c", f"printf {data_in} | {self.stub_name}").env(PATH=os.fsdecode(self.tmpdir))
        child = await self.thread.exec(command)
        self.nursery.start_soon(child.check)
        argv, new_thread = await self.server.accept()
        valid, _ = await new_thread.stdin.read(
            await new_thread.ram.malloc(bytes, len(data_in)))
        self.assertEqual(data_in, (await valid.read()).decode())
    



"""Async-library-independent shift/reset

Here we provide an abstraction for shift/reset, and resulting continuations,
which works identically across different Python async libraries. Our
implementation of shift/reset is single-shot and single-prompt.

One can call "shift" from any context, and it works identically.  "reset" also
works from any context, but it's a normal non-async function, so that's to be
expected.

Currently we just support trio and "native" continuations produced by calling
shift directly under reset.

As mentioned in the "dneio" module docstring, check out this tutorial if you
aren't familiar with shift/reset:
http://pllab.is.ocha.ac.jp/~asai/cw2011tutorial/main-e.pdf

"""

from __future__ import annotations
from dataclasses import dataclass
from dneio.outcome import Outcome
from trio._core._run import GLOBAL_RUN_CONTEXT, WaitTaskRescheduled, CancelShieldedCheckpoint, PermanentlyDetachCoroutineObject
import abc
import enum
import functools
import logging
import outcome
import trio
import types
import typing as t

__all__ = [
    'shift',
    'reset',
    'Continuation',
    'TrioSystemWaitReadable', 'set_trio_system_wait_readable',
]

logger = logging.getLogger(__name__)

SendType = t.TypeVar('SendType')
ReturnType = t.TypeVar('ReturnType')
AnswerType = t.TypeVar('AnswerType')
YieldType = t.TypeVar('YieldType')

class AnsweringContinuation(t.Generic[AnswerType, SendType, ReturnType]):
    """Something which can be passed SendType once, and produce AnswerType or ReturnType

    This is basically a single-shot function from SendType to ReturnType, except
    that we can also get a value of AnswerType, if the computation calls `shift`.
    Presumably, that `shift` either arranges to store the resulting continuation
    somewhere, or includes it in AnswerType so that the caller can deal with it.

    In practice, we mostly use Continuation, which has None for its answer and
    return types.

    """
    @abc.abstractmethod
    def resume(self, value: Outcome[SendType]) -> t.Union[AnswerType, ReturnType]: ...
    @abc.abstractmethod
    def send(self, value: SendType) -> t.Union[AnswerType, ReturnType]: ...
    @abc.abstractmethod
    def throw(self, exn: BaseException) -> t.Union[AnswerType, ReturnType]: ...
    # right now, everything is cancellable at all times.
    # we'll make this more customizable later.
    @abc.abstractmethod
    def is_cancelled(self) -> bool: ...

    def __call__(self, value: SendType) -> t.Union[AnswerType, ReturnType]:
        return self.send(value)

Continuation = AnsweringContinuation[None, SendType, None]
"""A Continuation whose return and answer types are both None

This is a degenerate form, which is most useful for callbacks registed on some
object, which shouldn't be returning or answering anything to the task invoking
the callback.

Really, this should be called NoneContinuation or something, and
AnswerContinuation should be called Continuation, but this is the more common
form, so it's nice for it to have the shorter name.

"""

class Runner(enum.Enum):
    TRIO = "trio"
    NATIVE = "native"

# For now, we assume we start under trio
_under_coro_runner: Runner = Runner.TRIO

def is_running_directly_under_trio() -> bool:
    return _under_coro_runner == Runner.TRIO


#### Implementation of "native" continuations
class Shift(t.Generic[AnswerType, SendType, ReturnType]):
    "The internal type we yield up to implement the shift function for native coroutines"
    __slots__ = ('func')
    def __init__(self, func: t.Callable[
            [AnsweringContinuation[AnswerType, SendType, ReturnType]], AnswerType]) -> None:
        self.func = func

ShiftCoroutine = t.Coroutine[Shift[AnswerType, SendType, ReturnType], SendType, ReturnType]
"A coroutine which yields Shift. Basically, a 'native' dneio coroutine"

def reset(
        coro: ShiftCoroutine[AnswerType, SendType, ReturnType],
) -> t.Union[AnswerType, ReturnType]:
    "Run this coro such that it can call `shift` and receive its continuation"
    global _under_coro_runner
    try:
        previous_runner = _under_coro_runner
        _under_coro_runner = Runner.NATIVE
        yielded_value: Shift[AnswerType, SendType, ReturnType]
        yielded_value = coro.send(None) # type: ignore
        if not isinstance(yielded_value, Shift):
            coro.throw(TypeError, TypeError("no yielding non-shifts!", yielded_value))
            raise TypeError("coro", coro, "yielded something other than a Shift")
    except StopIteration as e:
        return_value: ReturnType = e.value
        return return_value
    finally:
        _under_coro_runner = previous_runner
    answer_value: AnswerType = yielded_value.func(NativeContinuation(coro))
    return answer_value

@dataclass
class NativeContinuation(AnsweringContinuation[AnswerType, SendType, ReturnType]):
    __slots__ = ('coro')
    coro: ShiftCoroutine[AnswerType, SendType, ReturnType]

    def resume(self, value: Outcome[SendType]) -> t.Union[AnswerType, ReturnType]:
        if isinstance(value, outcome.Value):
            return self.send(value.value)
        else:
            return self.throw(value.error)

    # NativeContinuation.{send,throw} and reset are very similar; they could be
    # abstracted, but that would add stack frames, and keeping them separate is
    # an easy way to remove those stack frames, pending a C implementation.
    def send(self, value: SendType) -> t.Union[AnswerType, ReturnType]:
        global _under_coro_runner
        try:
            previous_runner = _under_coro_runner
            _under_coro_runner = Runner.NATIVE
            yielded_value: Shift[AnswerType, SendType, ReturnType]
            yielded_value = self.coro.send(value)
        except StopIteration as e:
            return_value: ReturnType = e.value
            return return_value
        finally:
            _under_coro_runner = previous_runner
        if not isinstance(yielded_value, Shift):
            self.coro.throw(TypeError("no yielding non-shifts!", yielded_value)) # type: ignore
            raise TypeError("coro", self.coro, "yielded something other than a Shift")
        answer_value: AnswerType = yielded_value.func(NativeContinuation(self.coro))
        return answer_value

    def throw(self, exn: BaseException) -> t.Union[AnswerType, ReturnType]:
        global _under_coro_runner
        try:
            previous_runner = _under_coro_runner
            _under_coro_runner = Runner.NATIVE
            yielded_value: Shift[AnswerType, SendType, ReturnType]
            yielded_value = self.coro.throw(exn) # type: ignore
        except StopIteration as e:
            return_value: ReturnType = e.value
            return return_value
        finally:
            _under_coro_runner = previous_runner
        if not isinstance(yielded_value, Shift):
            self.coro.throw(TypeError("no yielding non-shifts!", yielded_value)) # type: ignore
            raise TypeError("coro", self.coro, "yielded something other than a Shift")
        answer_value: AnswerType = yielded_value.func(NativeContinuation(self.coro))
        return answer_value

    def is_cancelled(self) -> bool:
        return False


#### Implementation of trio continuations
TrioTask = t.Any
"trio doesn't expose the type of Task publicly..."

@dataclass
class TrioContinuation(Continuation[SendType]):
    # We inherit from Continuation not AnsweringContinuation because it's
    # difficult to assign a proper AnswerType to trio tasks - easier for now to just
    # flatly say "None".
    __slots__ = ('task', 'cancelled', 'saved_send')
    task: TrioTask
    cancelled: bool
    saved_send: t.Optional[Outcome[None]]
    on_stack: bool

    # This essentially repeats a large part of the trio run loop. It would be
    # nicer if trio exposed a primtive for this directly.
    def resume(self, value: Outcome[SendType]) -> None:
        if self.cancelled:
            # discard the result - not great, obviously...
            logger.debug("TrioContinuation(%s): resumed after cancellation", self.task)
            return
        if self.on_stack:
            logger.debug("TrioContinuation(%s): immediately resumed with %s", self.task, value)
            # This will happen if the function passed to shift immediately resumes the
            # continuation. With trio, we run the function passed to shift on the
            # coroutine that's being suspended. So we can't resume the coroutine here,
            # since it's already running. Instead we'll save the outcome, and in shift()
            # we check saved_send and just return immediately if it's set. This is not
            # normal shift/reset semantics but it's the best we can do with how trio is
            # structured.
            self.saved_send = value
            return
        resuming_task = GLOBAL_RUN_CONTEXT.task
        runner = GLOBAL_RUN_CONTEXT.runner
        logger.debug("TrioContinuation(%s): resuming with %s", self.task, value)
        global _under_coro_runner
        try:
            previous_runner = _under_coro_runner
            _under_coro_runner = Runner.TRIO
            # We have to temporarily set GLOBAL_RUN_CONTEXT.task to the task that is being
            # resumed; after all, that's the task that's really going to be running. This
            # wouldn't be necessary if we had proper dynamically scoped variables in
            # Python :(
            GLOBAL_RUN_CONTEXT.task = self.task
            # a little bit of reschedule(), before we run the task
            self.task._abort_func = None
            self.task.custom_sleep_data = None
            try:
                msg = self.task.context.run(self.task.coro.send, value)
            except StopIteration as exn:
                logger.debug("TrioContinuation(%s): return %s", self.task, exn.value)
                GLOBAL_RUN_CONTEXT.runner.task_exited(self.task, outcome.Value(exn.value))
                return
            except BaseException as exn:
                logger.debug("TrioContinuation(%s): raised %s", self.task, exn)
                exn = exn.with_traceback(exn.__traceback__ and exn.__traceback__.tb_next)
                GLOBAL_RUN_CONTEXT.runner.task_exited(self.task, outcome.Error(exn))
                return
            logger.debug("TrioContinuation(%s): yield %s", self.task, msg)
        finally:
            _under_coro_runner = previous_runner
            GLOBAL_RUN_CONTEXT.task = resuming_task
        if msg is CancelShieldedCheckpoint:
            runner.reschedule(self.task)
        elif type(msg) is WaitTaskRescheduled:
            self.task._abort_func = msg.abort_func
            if runner.ki_pending and self.task is runner.main_task:
                self.task._attempt_delivery_of_pending_ki()
            self.task._attempt_delivery_of_any_pending_cancel()
        elif type(msg) is PermanentlyDetachCoroutineObject:
            runner.task_exited(self.task, msg.final_outcome)
        else:
            raise TypeError("bad yield from continuation", msg)

    def send(self, value: SendType) -> None:
        return self.resume(outcome.Value(value))

    def throw(self, exn: BaseException) -> None:
        return self.resume(outcome.Error(exn))

    def _abort_func(self, raise_cancel) -> trio.lowlevel.Abort:
        logger.debug("TrioContinuation(%s): cancelled", self.task)
        self.cancelled = True
        return trio.lowlevel.Abort.SUCCEEDED

    def is_cancelled(self) -> bool:
        return self.cancelled

@types.coroutine
def shift(func: t.Callable[[Continuation[SendType]], AnswerType]) -> t.Generator[t.Any, t.Any, SendType]:
    """Call `func` with our current continuation and block until that continuation is resumed.

    This is a coroutine function, just implemented synchronously because this is
    the only place we actually yield from.

    `func` gets a Continuation, not a full-fledged AnsweringContinuation, because
    that's easier to support for now. But in theory `func` should be
    AnsweringContinuation[AnswerType, SendType, ReturnType]

    """
    ensure_system_trio_task_running()
    if _under_coro_runner == Runner.TRIO:
        trio_cont = TrioContinuation[SendType](
            trio.lowlevel.current_task(), False, None,
            on_stack=True,
        )
        # There's no surrounding reset to run `func`, so we just run
        # it here and throw away the answer value.
        func(trio_cont)
        trio_cont.on_stack = False
        if trio_cont.saved_send:
            # the continuation was resumed immediately by func
            return trio_cont.saved_send.unwrap()
        return (yield WaitTaskRescheduled(trio_cont._abort_func)).unwrap()
    elif _under_coro_runner == Runner.NATIVE:
        return (yield Shift(func))
    else:
        raise Exception("running under unsupported coroutine runner")

#### Implementation of non-essential trio system task class
class TrioSystemWaitReadable:
    """Run "wait_readable" in a trio system task so we can avoid blocking trio tasks

    See rsyscall.epoller for how this is used.

    Frustratingly, we have to call .ensure_running on every call to `shift`,
    because trio has no means of autostarting a system task. There has to be a
    better way.

    Note that just calling ensure_running in .wait or .wait_cb doesn't work; a
    coroutine may have already called that in a previous call to trio.run and be
    blocked inside TSWR, waiting for the system task to resume it.

    Before settling on this approach we used a system where `shift` in a trio
    task would automatically open a nursery and make an object much like this
    one, which would automatically be passed down via dynamic scope and magic
    priority inheritance. It fell apart because Python doesn't have proper
    dynamic scope and maintaining the proper inheritance was too hard. That kind
    of implicit, untyped inheritance is against the core idea of dneio anyway;
    at least this object is very explicit and concrete; we don't use it through
    a global variable, we explicitly pass it around. We just need to find a
    better way to support it in trio.

    """
    def __init__(self, fd_number: int) -> None:
        self.fd_number = fd_number
        # we shouldn't ever get more than one waiter
        self._ops_in, self._ops_out = trio.open_memory_channel(1)
        self._run_running = False

    def ensure_running(self) -> None:
        if not self._run_running:
            self._run_running = True
            trio.lowlevel.spawn_system_task(self.run)

    def wait_cb(self, cb: Continuation[None]) -> None:
        self._ops_in.send_nowait(cb)

    async def wait(self) -> None:
        return await shift(self.wait_cb)

    async def run(self) -> None:
        try:
            while True:
                cb = await self._ops_out.receive()
                try:
                    await trio.lowlevel.wait_readable(self.fd_number)
                except:
                    self.wait_cb(cb)
                    raise
                try:
                    cb.send(None)
                except:
                    logger.exception("TSWR.run: callback raised exception")
                    raise
        finally:
            logger.debug("TSWR.run: system task exiting")
            self._run_running = False

_trio_system_wait_readable: t.Optional[TrioSystemWaitReadable] = None

def set_trio_system_wait_readable(tswr: TrioSystemWaitReadable) -> None:
    global _trio_system_wait_readable
    _trio_system_wait_readable = tswr

def ensure_system_trio_task_running() -> None:
    "Called in every call to shift"
    if _trio_system_wait_readable:
        _trio_system_wait_readable.ensure_running()



from __future__ import annotations

from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread, FileDescriptor, Pointer
from rsyscall.epoller import *
import trio
import outcome

from rsyscall.tests.utils import do_async_things
from rsyscall.near.sysif import SyscallInterface, Syscall
from rsyscall.sys.syscall import SYS
from dneio import RequestQueue, reset, Continuation
import typing as t

class DelayResultSysif(SyscallInterface):
    def __init__(self, sysif: SyscallInterface,
                 delay_queue: RequestQueue[t.Tuple[Syscall, outcome.Outcome[int]], None]) -> None:
        self.sysif = sysif
        self.delay_queue = delay_queue

    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        syscall = Syscall(number, arg1, arg2, arg3, arg4, arg5, arg6)
        ret = await outcome.acapture(self.sysif.syscall, number, arg1, arg2, arg3, arg4, arg5, arg6)
        await self.delay_queue.request((syscall, ret))
        return ret.unwrap()

    async def close_interface(self) -> None:
        return await self.sysif.close_interface()

    def get_activity_fd(self) -> t.Optional[FileDescriptor]:
        return self.sysif.get_activity_fd()

class TestEpoller(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test_local(self) -> None:
        await do_async_things(self, self.thr.epoller, self.thr)

    async def test_multi(self) -> None:
        await do_async_things(self, self.thr.epoller, self.thr, 0)
        async with trio.open_nursery() as nursery:
            for i in range(1, 6):
                nursery.start_soon(do_async_things, self, self.thr.epoller, self.thr, i)

    async def test_thread_multi(self) -> None:
        thread = await self.thr.clone()
        await do_async_things(self, thread.epoller, thread, 0)
        async with trio.open_nursery() as nursery:
            for i in range(1, 6):
                nursery.start_soon(do_async_things, self, thread.epoller, thread, i)

    async def test_thread_root_epoller(self) -> None:
        thread = await self.thr.clone()
        epoller = await Epoller.make_root(thread.ram, thread.task)
        await do_async_things(self, epoller, thread)

    async def test_afd_with_handle(self):
        pipe = await self.thr.pipe()
        afd = await self.thr.make_afd(pipe.write, set_nonblock=True)
        new_afd = afd.with_handle(pipe.write)
        await new_afd.write_all_bytes(b'foo')

    async def test_delayed_eagain(self):
        pipe = await self.thr.pipe()
        thread = await self.thr.clone()
        async_pipe_rfd = await thread.make_afd(thread.inherit_fd(pipe.read), set_nonblock=True)
        # write in parent, read in child
        input_data = b'hello'
        buf_to_write: Pointer[bytes] = await self.thr.ptr(input_data)
        buf_to_write, _ = await pipe.write.write(buf_to_write)
        self.assertEqual(await async_pipe_rfd.read_some_bytes(), input_data)
        buf = await thread.malloc(bytes, 4096)
        # set up the EAGAIN to be delayed
        queue: RequestQueue[t.Tuple[Syscall, outcome.Outcome[int]], None] = RequestQueue()
        old_sysif = thread.task.sysif
        thread.task.sysif = DelayResultSysif(old_sysif, queue)
        @self.nursery.start_soon
        async def race_eagain():
            # wait for EAGAIN
            (syscall, result), cb = await queue.get_one()
            self.assertIsInstance(result.error, BlockingIOError)
            thread.task.sysif = old_sysif
            queue.close(Exception("remaining syscalls?"))
            # write data after the EAGAIN
            await pipe.write.write(buf_to_write)
            # give epoll event a chance to be read - it will be available immediately
            await trio.sleep(0)
            # resume the suspended EAGAIN coroutine, which should keep running and get data
            cb.send(None)
        valid, remaining = await async_pipe_rfd.read(buf)

    async def test_wrong_op_on_pipe(self):
        "Reading or writing to the wrong side of a pipe fails immediately with an error"
        pipe = await self.thr.pipe()
        async_pipe_wfd = await self.thr.make_afd(pipe.write, set_nonblock=True)
        async_pipe_rfd = await self.thr.make_afd(pipe.read, set_nonblock=True)
        # we actually are defined to get EBADF in this case, which is
        # a bit of a worrying error, but whatever
        with self.assertRaises(OSError) as cm:
            await async_pipe_wfd.read_some_bytes()
        self.assertEqual(cm.exception.errno, 9)
        with self.assertRaises(OSError) as cm:
            await async_pipe_rfd.write_all_bytes(b'hi')
        self.assertEqual(cm.exception.errno, 9)



"`#include <signal.h>`"
from __future__ import annotations
import typing as t
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.struct import Struct, bits
from dataclasses import dataclass, field
import enum
import contextlib
import rsyscall.near.types as near
from rsyscall.far import Task
from rsyscall.handle import Pointer, WrittenPointer

import signal

__all__ = [
    "SIG",
    "SA",
    "HowSIG",
    "Sighandler",
    "Siginfo",
    "Sigset",
    "Sigaction",
    "SignalBlock",
    "SignalTask",
]

class SIG(enum.IntEnum):
    "All the non-realtime signals, used by many syscalls"
    NONE = 0
    HUP = signal.SIGHUP
    INT = signal.SIGINT
    QUIT = signal.SIGQUIT
    ILL = signal.SIGILL
    TRAP = signal.SIGTRAP
    ABRT = signal.SIGABRT
    IOT = signal.SIGIOT
    BUS = signal.SIGBUS
    FPE = signal.SIGFPE
    KILL = signal.SIGKILL
    USR1 = signal.SIGUSR1
    SEGV = signal.SIGSEGV
    USR2 = signal.SIGUSR2
    PIPE = signal.SIGPIPE
    ALRM = signal.SIGALRM
    TERM = signal.SIGTERM
    STKFLT = lib.SIGSTKFLT
    CHLD = signal.SIGCHLD
    CONT = signal.SIGCONT
    STOP = signal.SIGSTOP
    TSTP = signal.SIGTSTP
    TTIN = signal.SIGTTIN
    TTOU = signal.SIGTTOU
    URG = signal.SIGURG
    XCPU = signal.SIGXCPU
    XFSZ = signal.SIGXFSZ
    VTALRM = signal.SIGVTALRM
    PROF = signal.SIGPROF
    WINCH = signal.SIGWINCH
    IO = signal.SIGIO
    PWR = signal.SIGPWR
    SYS = signal.SIGSYS

class SA(enum.IntFlag):
    "Flags which can be set in a struct sigaction passed to the sigaction syscall"
    NOCLDSTOP = lib.SA_NOCLDSTOP
    NOCLDWAIT = lib.SA_NOCLDWAIT
    NODEFER = lib.SA_NODEFER
    ONSTACK = lib.SA_ONSTACK
    RESETHAND = lib.SA_RESETHAND
    RESTART = lib.SA_RESTART
    SIGINFO = lib.SA_SIGINFO
    RESTORER = lib.SA_RESTORER

class HowSIG(enum.IntEnum):
    "The how argument to sigprocmask"
    BLOCK = lib.SIG_BLOCK
    UNBLOCK = lib.SIG_UNBLOCK
    SETMASK = lib.SIG_SETMASK

class Sighandler(enum.IntEnum):
    "Special-cased signal handler values used by sigaction"
    IGN = signal.Handlers.SIG_IGN
    DFL = signal.Handlers.SIG_DFL

@dataclass
class Siginfo(Struct):
    "struct siginfo, returned by many syscalls"
    code: int
    pid: int
    uid: int
    status: int

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct siginfo*')
        struct.si_code = self.code
        struct.si_pid = self.pid
        struct.si_uid = self.uid
        struct.si_status = self.status
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='Siginfo')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct siginfo const*', ffi.from_buffer(data))
        return cls(
            code=struct.si_code,
            pid=struct.si_pid,
            uid=struct.si_uid,
            status=struct.si_status,
        )

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct siginfo')

class Sigset(t.Set[SIG], Struct):
    """A fixed-size 64-signal struct sigset

    Technically, at the syscall level, struct sigset is variable size. It was
    made variable size when the number of signals was increased from 32 to 64
    and the "rt" variants of many signal syscalls was added, taking as an
    additional argument the size of struct sigset. But it's unlikely the number
    of signals will ever be increased again on Linux, and even if that does
    happen, it will be backwrds-compatible.

    """
    def to_cffi(self) -> t.Any:
        set_integer = 0
        for sig in self:
            set_integer |= 1 << (sig-1)
        return ffi.new('struct kernel_sigset*', (set_integer,))

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(self.to_cffi()))

    T = t.TypeVar('T', bound='Sigset')
    @classmethod
    def from_cffi(cls: t.Type[T], struct: t.Any) -> T:
        return cls({SIG(bit) for bit in bits(struct.val)})

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct kernel_sigset*', ffi.from_buffer(data))
        return cls.from_cffi(struct)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct kernel_sigset')

@dataclass
class Sigaction(Struct):
    "struct sigaction, passed to and returned from the sigaction syscall"
    handler: t.Union[Sighandler, near.Address]
    flags: SA = SA(0)
    mask: Sigset = field(default_factory=Sigset)
    restorer: t.Optional[near.Address] = None

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct kernel_sigaction const*', {
            "ksa_handler": ffi.cast('sighandler_t', int(self.handler)),
            "ksa_flags": self.flags,
            "ksa_restorer": ffi.cast('sigrestore_t', int(self.restorer or 0)),
            "ksa_mask": self.mask.to_cffi()[0],
        })))

    T = t.TypeVar('T', bound='Sigaction')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct kernel_sigaction const*', ffi.from_buffer(data))
        handler: t.Union[Sighandler, near.Address]
        int_handler = int(ffi.cast('long int', struct.ksa_handler))
        try:
            handler = Sighandler(int_handler)
        except ValueError:
            handler = near.Address(int_handler)
        int_restorer = int(ffi.cast('long int', struct.ksa_restorer))
        return cls(
            handler=handler,
            flags=SA(struct.ksa_flags),
            mask=Sigset.from_cffi(struct.ksa_mask),
            restorer=near.Address(int_restorer) if int_restorer else None,
        )

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct kernel_sigaction')



#### Classes ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _rt_sigprocmask(sysif: SyscallInterface,
                          newset: t.Optional[t.Tuple[HowSIG, near.Address]],
                          oldset: t.Optional[near.Address],
                          sigsetsize: int) -> None:
    "The raw, near, rt_sigprocmask syscall."
    if newset is not None:
        how, set = newset
    else:
        how, set = 0, 0 # type: ignore
    if oldset is None:
        oldset = 0 # type: ignore
    await sysif.syscall(SYS.rt_sigprocmask, how, set, oldset, sigsetsize)

class SignalTask(Task):
    "The subset of functionality of the Task related to blocking and handling signals"
    def __post_init__(self) -> None:
        super().__post_init__()
        self.sigmask = Sigset()
        self.old_sigmask = Sigset()
        self.sigmask_oldset_ptr: t.Optional[Pointer[Sigset]] = None
        self.sigprocmask_running = False

    async def sigaction(self, signum: SIG,
                        act: t.Optional[Pointer[Sigaction]],
                        oldact: t.Optional[Pointer[Sigaction]]) -> None:
        with contextlib.ExitStack() as stack:
            act_n = self._borrow_optional(stack, act)
            oldact_n = self._borrow_optional(stack, oldact)
            # rt_sigaction takes the size of the sigset, not the size of the sigaction;
            # and sigset is a fixed size.
            await _rt_sigaction(self.sysif, signum, act_n, oldact_n, Sigset.sizeof())

    async def _sigprocmask(self, newset: t.Optional[t.Tuple[HowSIG, WrittenPointer[Sigset]]],
                           oldset: t.Optional[Pointer[Sigset]]=None) -> t.Optional[Pointer[Sigset]]:
        "The low-level implementation of sigprocmask, without any tracking of our current sigmask."
        with contextlib.ExitStack() as stack:
            newset_n: t.Optional[t.Tuple[HowSIG, near.Address]]
            if newset:
                newset_ptr_n = stack.enter_context(newset[1].borrow(self))
                newset_n = newset[0], newset_ptr_n
            else:
                newset_n = None
            oldset_n = self._borrow_optional(stack, oldset)
            await _rt_sigprocmask(self.sysif, newset_n, oldset_n, Sigset.sizeof())
        if oldset:
            return oldset
        else:
            return None

    @t.overload
    async def sigprocmask(self, newset: t.Tuple[HowSIG, WrittenPointer[Sigset]]) -> None: ...
    @t.overload
    async def sigprocmask(self, newset: t.Optional[t.Tuple[HowSIG, WrittenPointer[Sigset]]],
                          oldset: Pointer[Sigset]) -> Pointer[Sigset]: ...

    async def sigprocmask(self, newset: t.Optional[t.Tuple[HowSIG, WrittenPointer[Sigset]]],
                          oldset: t.Optional[Pointer[Sigset]]=None) -> t.Optional[Pointer[Sigset]]:
        "sigprocmask, with additional tracking of what we believe our sigmask is"
        if self.sigprocmask_running:
            # TODO we can remove this restriction if we can determine the real order that
            # syscalls are made, so we can accurately maintain sigmask; that's currently
            # tricky due to pipelining.
            raise Exception("sigprocmask is currently being called, "
                            "we disallow multiple simultaneous calls for implementation simplicity")
        self.sigprocmask_running = True
        if newset is None:
            ret = await self._sigprocmask(None, oldset)
            self.sigprocmask_running = False
            return ret
        else:
            how, set = newset
            if how == HowSIG.BLOCK:
                new_sigmask = Sigset(self.sigmask.union(set.value))
            elif how == HowSIG.UNBLOCK:
                new_sigmask = Sigset(self.sigmask - set.value)
            elif how == HowSIG.SETMASK:
                new_sigmask = Sigset(set.value)
            ret = await self._sigprocmask(newset, oldset)
            self.sigprocmask_running = False
            if oldset is not None:
                self.old_sigmask = self.sigmask
                self.sigmask_oldset_ptr = oldset
            self.sigmask = new_sigmask
            return ret

    async def sigmask_block(self, newset: WrittenPointer[Sigset], oldset: t.Optional[Pointer[Sigset]]=None) -> SignalBlock:
        "Block some signals and get back a SignalBlock to witness and own those blocked signals"
        if len(newset.value.intersection(self.sigmask)) != 0:
            raise Exception("can't allocate a SignalBlock for a signal that was already blocked",
                            newset.value, self.sigmask)
        if oldset:
            await self.sigprocmask((HowSIG.BLOCK, newset), oldset)
        else:
            await self.sigprocmask((HowSIG.BLOCK, newset))
        return SignalBlock(self, newset)

    async def read_oldset_and_check(self) -> None:
        """Read the most recent oldset buffer passed to sigprocmask and check it matches what we expect

        We do this in a separate method to avoid forcing the user to read memory
        if they don't want to validate that the oldset buffer contains what they
        expect. They can call it later if they want to do the check.

        """
        if self.sigmask_oldset_ptr is None:
            raise Exception("can't check our tracking of sigmask "
                            "when we haven't called sigprocmask with an oldset ptr")
        sigmask = await self.sigmask_oldset_ptr.read()
        if (sigmask & self.old_sigmask) != self.old_sigmask:
            raise Exception("SignalMask tracking got out of sync, we blocked these signals:", self.old_sigmask,
                            "but the actually blocked signals don't include all of those:", sigmask)

@dataclass(eq=False)
class SignalBlock:
    """This represents some signals being blocked from normal handling

    We need this around to use alternative signal handling mechanisms
    such as signalfd.

    This serves as both a witness that we have indeed blocked these signals, and
    an owner to help us unblock those signals if and when we no longer need to
    block them.

    """
    task: SignalTask
    newset: WrittenPointer[Sigset]

    @property
    def mask(self) -> Sigset:
        return self.newset.value

    async def unblock(self, oldset: t.Optional[Pointer[Sigset]]=None) -> None:
        if oldset:
            await self.task.sigprocmask((HowSIG.UNBLOCK, self.newset), oldset)
        else:
            await self.task.sigprocmask((HowSIG.UNBLOCK, self.newset))


#### Raw syscalls ####
async def _rt_sigaction(sysif: SyscallInterface, signum: SIG,
                        act: t.Optional[near.Address],
                        oldact: t.Optional[near.Address],
                        size: int) -> None:
    if act is None:
        act = 0 # type: ignore
    if oldact is None:
        oldact = 0 # type: ignore
    await sysif.syscall(SYS.rt_sigaction, signum, act, oldact, size)

async def _kill(sysif: SyscallInterface, pid: t.Union[near.Process, near.ProcessGroup], sig: SIG) -> None:
    if isinstance(pid, near.ProcessGroup):
        pid = -int(pid) # type: ignore
    await sysif.syscall(SYS.kill, pid, sig)


#### Tests ####
from unittest import TestCase

class TestSignal(TestCase):
    def test_siginfo(self) -> None:
        initial = Siginfo(13, 581, 1092, 12309)
        output = Siginfo.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)

    def test_sigaction(self) -> None:
        sa = Sigaction(Sighandler.IGN, SA(0), Sigset(), near.Address(0x42))
        out_sa = Sigaction.from_bytes(sa.to_bytes())
        self.assertEqual(sa.handler, out_sa.handler)
        self.assertEqual(sa.flags, out_sa.flags)
        self.assertEqual(sa.mask, out_sa.mask)
        self.assertEqual(sa.restorer, out_sa.restorer)

        sa = Sigaction(Sighandler.DFL, SA.RESTART|SA.RESETHAND,
                       Sigset({SIG.INT, SIG.TERM}),
                       near.Address(0))
        out_sa = Sigaction.from_bytes(sa.to_bytes())
        self.assertEqual(sa.handler, out_sa.handler)
        self.assertEqual(sa.flags, out_sa.flags)
        self.assertEqual(sa.mask, out_sa.mask)
        self.assertEqual(None, out_sa.restorer)



"`#include <sys/eventfd.h>`"
from __future__ import annotations
from rsyscall._raw import lib # type: ignore
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS
import enum
import rsyscall.near.types as near
import typing as t
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask

__all__ = [
    "EFD",
    "EventFileDescriptor",
    "EventfdTask",
]

class EFD(enum.IntFlag):
    NONE = 0
    CLOEXEC = lib.EFD_CLOEXEC
    NONBLOCK = lib.EFD_NONBLOCK
    SEMAPHORE = lib.EFD_SEMAPHORE

async def _eventfd(sysif: SyscallInterface, initval: int, flags: EFD) -> near.FileDescriptor:
    "The raw, near, eventfd syscall."
    return near.FileDescriptor(await sysif.syscall(SYS.eventfd2, initval, flags))

T_fd = t.TypeVar('T_fd', bound='EventFileDescriptor')
class EventFileDescriptor(BaseFileDescriptor):
    pass

class EventfdTask(FileDescriptorTask[T_fd]):
    async def eventfd(self, initval: int, flags: EFD=EFD.NONE) -> T_fd:
        return self.make_fd_handle(await _eventfd(self.sysif, initval, flags|EFD.CLOEXEC))



"""A pure asynchronous REPL, capable of parsing and running async code given as input

PureREPL is a basic async REPL. It's pure in the sense that it doesn't itself read input;
the user must pass input to it as strings.

"""
from dataclasses import dataclass
import ast
import traceback
import typeguard
import typing as t
from arepl.astcodeop import ast_compile_interactive
from arepl.aeval import (
    ReturnResult, ExceptionResult, ExpressionResult, FallthroughResult, Result, eval_single,
)
from arepl.help import help_to_str

class FromREPL(Exception):
    def __init__(self, exn: Exception) -> None:
        self.exn = exn

class PureREPL:
    def __init__(self, global_vars: t.Dict[str, t.Any]) -> None:
        self.global_vars = global_vars
        self.buf = ""

    async def eval_single(self, astob: ast.Interactive) -> Result:
        return (await eval_single(astob, self.global_vars))

    async def add_line(self, data: str) -> t.Optional[Result]:
        """Add a single line to the REPL buffer and try to parse and evaluate the AST.

        If no AST can be parsed from the buffer, returns None.
        
        Make sure to pass exactly one single line to this function,
        including newline! Otherwise, you might pass multiple
        statements at once, and the parser won't like that.

        """
        self.buf += data
        try:
            # remove the last newline
            astob = ast_compile_interactive(self.buf[:-1])
        except Exception:
            self.buf = ""
            raise
        else:
            if astob is not None:
                self.buf = ""
                return (await self.eval_single(astob))
            return None

class LineBuffer:
    "A simple character buffer to split data into lines"
    def __init__(self) -> None:
        self.buf: str = ""

    def add(self, data: str) -> t.List[str]:
        self.buf += data
        *lines, self.buf = self.buf.split('\n')
        return [line + '\n' for line in lines]

# TODO I should also be able to pass in a predicate function which I call on the return value.
# That way I can represent constraints on the returned value at a value level.
# (I still do the wanted_type so that mypy type checking is correct)
T = t.TypeVar('T')
async def run_repl(read: t.Callable[[], t.Awaitable[bytes]],
                   write: t.Callable[[bytes], t.Awaitable[None]],
                   global_vars: t.Dict[str, t.Any], wanted_type: t.Type[T]) -> T:
    async def print_to_user(*args) -> None:
        await write((" ".join([str(arg) for arg in args]) + "\n").encode())
    async def help_to_user(request) -> None:
        await write(help_to_str(request).encode())
    async def print_exn(e: BaseException) -> None:
        # this call to run_repl may take place at a time where an exception is being handled; when
        # we print this internal exception, we don't want to print that outside exception context,
        # since it's irrelevant.
        e.__suppress_context__ = True # type: ignore
        await write("".join(traceback.format_exception(None, e, e.__traceback__)).encode())
    global_vars['print'] = print_to_user
    global_vars['help'] = help_to_user
    repl = PureREPL(global_vars)
    line_buf = LineBuffer()
    await write(b">")
    while True:
        raw_data = await read()
        if len(raw_data) == 0:
            raise Exception("REPL hangup")
        for line in line_buf.add(raw_data.decode()):
            try:
                result = await repl.add_line(line)
            except Exception as exn:
                await print_exn(exn)
                continue
            if result is None:
                continue
            if isinstance(result, ReturnResult):
                try:
                    typeguard.check_type('return value', result.value, wanted_type)
                except TypeError as e:
                    await print_exn(e)
                else:
                    return result.value
            elif isinstance(result, ExceptionResult):
                if isinstance(result.exception, FromREPL):
                    raise result.exception
                else:
                    await print_exn(result.exception)
            elif isinstance(result, ExpressionResult):
                await print_to_user(result.value)
                global_vars['_'] = result.value
            elif isinstance(result, FallthroughResult):
                pass
            else:
                raise Exception("bad Result returned from PureREPL", result)
            await write(b">")



"""Make a new thread by calling clone

That is to say, make a new thread the normal way.

"""
from __future__ import annotations
from dneio import reset
from dataclasses import dataclass
from rsyscall._raw import ffi # type: ignore
from rsyscall.epoller import AsyncFileDescriptor
from rsyscall.handle import Stack, WrittenPointer, Pointer, FutexNode, FileDescriptor, Task, FutexNode
from rsyscall.loader import Trampoline, NativeLoader
from rsyscall.memory.allocator import Arena
from rsyscall.memory.ram import RAM
from rsyscall.monitor import AsyncChildProcess, ChildProcessMonitor
from rsyscall.network.connection import Connection
from rsyscall.struct import Int32
from rsyscall.tasks.connection import SyscallConnection
from rsyscall.near.sysif import SyscallError
import contextlib
import logging
import rsyscall.far as far
import rsyscall.handle as handle
import trio
import typing as t

from rsyscall.sched import CLONE
from rsyscall.signal import SIG
from rsyscall.sys.mman import PROT, MAP
from rsyscall.sys.socket import SHUT
from rsyscall.sys.wait import W

__all__ = [
    'launch_futex_monitor',
    'clone_child_task',
]

logger = logging.getLogger(__name__)

async def launch_futex_monitor(ram: RAM,
                               loader: NativeLoader, monitor: ChildProcessMonitor,
                               futex_pointer: WrittenPointer[FutexNode]) -> AsyncChildProcess:
    """Launch a process to wait on a futex; then we monitor the process to monitor the futex

    This process calls futex(futex_pointer, FUTEX_WAIT, futex_pointer.value) and
    then exits, so this process will exit if and when the futex has FUTEX_WAKE
    called on it.

    Sadly, this is the best we can do with integrating futexes into our event
    loop. There used to be a way to get a file descriptor to represent a futex,
    but it was removed because it was racy.

    Something better would be really great - especially because it would allow
    incorporating pthreads locks and other shared memory concurrency mechanisms
    based on futexes, into a normal event loop.

    """
    async def op(sem: RAM) -> t.Tuple[Pointer[Stack], WrittenPointer[Stack]]:
        stack_value = loader.make_trampoline_stack(Trampoline(
            loader.futex_helper_func, [
                int(futex_pointer.near + ffi.offsetof('struct futex_node', 'futex')),
                futex_pointer.value.futex]))
        stack_buf = await sem.malloc(Stack, 4096)
        stack = await stack_buf.write_to_end(stack_value, alignment=16)
        return stack
    stack = await ram.perform_batch(op)
    futex_process = await monitor.clone(CLONE.VM|CLONE.FILES, stack)
    # wait for futex helper to SIGSTOP itself,
    # which indicates the trampoline is done and we can deallocate the stack.
    state = await futex_process.waitpid(W.EXITED|W.STOPPED)
    if state.state(W.EXITED):
        raise Exception("thread internal futex-waiting task died unexpectedly", state)
    # resume the futex_process so it can start waiting on the futex
    await futex_process.kill(SIG.CONT)
    # TODO uh we need to actually call something to free the stack
    return futex_process

async def clone_child_task(
        task: Task,
        ram: RAM,
        connection: Connection,
        loader: NativeLoader,
        monitor: ChildProcessMonitor,
        flags: CLONE,
        trampoline_func: t.Callable[[FileDescriptor], Trampoline],
) -> t.Tuple[AsyncChildProcess, Task]:
    """Clone a new child process and setup the sysif and task to manage it

    We rely on trampoline_func to take a socket and give us a native function call with
    arguments that will speak the rsyscall protocol over that socket.

    We want to see EOF on our local socket if that remote socket is no longer being read;
    for example, if the process exits or execs.
    This is not automatic for us: Since the process might share its file descriptor table
    with other processes, remote_sock might not be closed when the process exits or execs.

    To ensure that we get an EOF, we use the ctid futex, which, thanks to
    CLONE.CHILD_CLEARTID, will be cleared and receive a futex wakeup when the child
    process exits or execs.

    When we see that futex wakeup (from Python, with the futex integrated into our event
    loop through launch_futex_monitor), we call shutdown(SHUT.RDWR) on the local socket
    from the parent. This results in future reads returning EOF.

    """
    # These flags are mandatory; if we don't use CLONE_VM then CHILD_CLEARTID doesn't work
    # properly and our only other recourse to detect exec is to abuse robust futexes.
    flags |= CLONE.VM|CLONE.CHILD_CLEARTID
    # Open a channel which we'll use for the rsyscall connection
    [(access_sock, remote_sock)] = await connection.open_async_channels(1)
    # Create a trampoline that will start the new process running an rsyscall server
    trampoline = trampoline_func(remote_sock)
    # TODO it is unclear why we sometimes need to make a new mapping here, instead of
    # allocating with our normal allocator; all our memory is already MAP.SHARED, I think.
    # We should resolve this so we can use the normal allocator.
    arena = Arena(await task.mmap(4096*2, PROT.READ|PROT.WRITE, MAP.SHARED))
    async def op(sem: RAM) -> t.Tuple[t.Tuple[Pointer[Stack], WrittenPointer[Stack]],
                                                       WrittenPointer[FutexNode]]:
        stack_value = loader.make_trampoline_stack(trampoline)
        stack_buf = await sem.malloc(Stack, 4096)
        stack = await stack_buf.write_to_end(stack_value, alignment=16)
        futex_pointer = await sem.ptr(FutexNode(None, Int32(1)))
        return stack, futex_pointer
    # Create the stack we'll need, and the zero-initialized futex
    stack, futex_pointer = await ram.perform_batch(op, arena)
    # it's important to start the processes in this order, so that the thread
    # process is the first process started; this is relevant in several
    # situations, including unshare(NEWPID) and manipulation of ns_last_pid
    child_process = await monitor.clone(flags, stack, ctid=futex_pointer)
    # We want to be able to rely on getting an EOF if the other side of the syscall
    # connection is no longer being read (e.g., if the process exits or execs).  Since the
    # process might share its file descriptor table with other processes, remote_sock
    # might not be closed when the process exits or execs. To ensure that we get an EOF,
    # we use the ctid futex, which will be cleared on process exit or exec; we shutdown
    # access_sock when the ctid futex is cleared, to get an EOF.
    # We do this with launch_futex_monitor and a background coroutine.
    futex_process = await launch_futex_monitor(ram, loader, monitor, futex_pointer)
    async def shutdown_access_sock_on_futex_process_exit():
        try:
            await futex_process.waitpid(W.EXITED)
        except SyscallError:
            # if the parent of the futex_process dies, this syscall
            # connection is broken anyway, so shut it down.
            pass
        await access_sock.handle.shutdown(SHUT.RDWR)
    # Running this in the background, without an associated object, is a bit dubious...
    reset(shutdown_access_sock_on_futex_process_exit())
    # Set up the new task with appropriately inherited namespaces, tables, etc.
    # TODO correctly track all the namespaces we're in
    if flags & CLONE.NEWPID:
        pidns = far.PidNamespace(child_process.process.near.id)
    else:
        pidns = task.pidns
    if flags & CLONE.FILES:
        fd_table = task.fd_table
    else:
        fd_table = handle.FDTable(child_process.process.near.id, task.fd_table)
    child_task = Task(
        child_process.process, fd_table, task.address_space, pidns)
    child_task.sigmask = task.sigmask
    # Move ownership of the remote sock into the task and store it so it isn't closed
    remote_sock_handle = remote_sock.inherit(child_task)
    await remote_sock.invalidate()
    # Create the new syscall interface, which needs to use not just the connection,
    # but also the futex process.
    child_task.sysif = SyscallConnection(
        logger.getChild(str(child_process.process.near)),
        access_sock, access_sock,
        remote_sock_handle, remote_sock_handle,
    )
    return child_process, child_task



"""Fundamental FD ownership and lifecycle

We should probably move this to unistd.fd; after all, we define the
_close system call in this module, as part of the fundamental
lifecycle of an FD.

"""

from __future__ import annotations
from dataclasses import dataclass
from weakref import WeakSet
import itertools
import abc
import gc
import rsyscall.far
import rsyscall.near
from rsyscall.near.sysif import SyscallHangup, SyscallInterface
from rsyscall.sys.syscall import SYS
import trio
import typing as t
import logging
import contextlib
logger = logging.getLogger(__name__)

from rsyscall.sched import CLONE, _unshare

T_fd = t.TypeVar('T_fd', bound='BaseFileDescriptor')
@dataclass(eq=False)
class BaseFileDescriptor:
    """A file descriptor accessed through some `Task`

    This is an rsyscall-internal base class,
    which other `FileDescriptor` objects inherit from for core lifecycle methods.
    See `FileDescriptor` for more information.

    """
    __slots__ = ('task', 'near', 'valid')
    task: FileDescriptorTask
    near: rsyscall.near.FileDescriptor
    valid: bool

    def _validate(self) -> None:
        if not self.valid:
            raise Exception("handle is no longer valid")

    def _invalidate(self) -> bool:
        """Invalidate this reference to this file descriptor

        Returns true if we removed the last reference, and are now responsible for closing the FD.

        """
        if self.valid:
            self.valid = False
            handles = self._remove_from_tracking()
            return len(handles) == 0
        else:
            return False

    async def invalidate(self) -> bool:
        """Invalidate this reference to this file descriptor, closing it if necessary

        Returns true if we removed the last reference, and closed the FD.

        We'll use the task inside the last file descriptor to be invalidated to actually
        do the close.
        """
        if self._invalidate():
            # we were the last handle for this fd, we should close it
            logger.debug("invalidating %s, no handles remaining, closing", self)
            fd_table = self.task.fd_table
            del fd_table.near_to_handles[self.near]
            await fd_table._close_fd(self.task, self.near)
            return True
        else:
            logger.debug("invalidating %s, some handles remaining", self)
            return False

    async def close(self) -> None:
        """Close this file descriptor if it's the only handle to it; throwing if there's other handles

        manpage: close(2)
        """
        if not self.is_only_handle():
            raise Exception("can't close this fd", self, "there are handles besides this one to it",
                            self._get_global_handles())
        if not self.valid:
            raise Exception("can't close an invalid FD handle")
        closed = await self.invalidate()
        if not closed:
            raise Exception("for some reason, the fd wasn't closed; "
                            "maybe some race condition where there are still handles left around?")

    def inherit(self, task: FileDescriptorTask[T_fd]) -> T_fd:
        """Make another FileDescriptor referencing the same file but using `task`, which inherited this FD, for syscalls

        Whenever we call `clone` (without passing `CLONE.FILES`),
        the file descriptors in the parent process are copied to the child process;
        this is tracked in rsyscall, and we can call `inherit` on any parent `FileDescriptor`
        to get a handle for the copy in the child.

        """
        return task.inherit_fd(self)

    def for_task(self, task: FileDescriptorTask[T_fd]) -> T_fd:
        """Make a new handle for the same FD, but using `task`, in the same FD table, for syscalls

        Two tasks in the same file descriptor table can be created by calling `clone` with `CLONE.FILES`.

        Once we call this method, we'll have multiple handles for a single file descriptor,
        in a single file descriptor table.
        We won't be able to use `close` to close the FD, since that would break other handles;
        we'll need to use `invalidate` instead.
        (Or, we can just rely on garbage collection.)

        If we want to access the file from another task, we may call the for_task method on
        the FileDescriptor, passing the other task from which we want to access the file.
        This will return another FileDescriptor referencing that file.  This will only work if
        the two tasks are in the same file descriptor table; that is typically the case for
        most scenarios and most kinds of threads. .

        """
        self._validate()
        if self.task.fd_table != task.fd_table:
            raise rsyscall.far.FDTableMismatchError(self.task.fd_table, task.fd_table)
        return task._make_fd_handle_from_near(self.near)

    @contextlib.contextmanager
    def borrow(self, task: FileDescriptorTask) -> t.Iterator[rsyscall.near.FileDescriptor]:
        "Validate that this FD can be accessed from this Task, and yield the near.FD to use for syscalls"
        # TODO we should be the only means of getting FD.near
        # TODO we should just set an in_use flag or something
        # oh argh, what about borrow_with, though?
        # hmm that's fine I guess... there's references inside...
        # ok, the thing is, we already can't move fds or pointers around
        # because we have references in memory
        # maybe borrowing should be another, more strong reference?
        # well, the point of this that we won't be freed during a syscall
        if self.task == task:
            yield self.near
        else:
            # note that we can't immediately change this to not use for_task,
            # because we need to get an FD which stays in the same fd table as task,
            # even if the task owning the FD we're borrowing switches fd tables
            borrowed = self.for_task(task)
            try:
                yield borrowed.near
            finally:
                # we can't call invalidate since we can't actually close this fd since that would
                # require more syscalls. we should really make it so that if the user tries to
                # invalidate the fd they passed into a syscall, they get an exception when they call
                # invalidate. but in lieu of that, we'll throw here. this will cause us to drop
                # events from syscalls, which would break a system that wants to handle exceptions
                # and resume, so we should fix this later. TODO
                # hmm actually I think it might be fine to borrow an fd and free its original?
                # that will happen if we borrow an expression... which should be fine...
                # maybe borrow is a bad design.
                # maybe borrow should just mean, you can't invalidate this fd right now.
                # though we do want to also check that it's the right address space...
                if borrowed.valid:
                    borrowed.valid = False
                    if len(borrowed._remove_from_tracking()) == 0:
                        raise Exception("borrowed fd must have been freed from under us, %s", borrowed)

    def move(self, task: FileDescriptorTask[T_fd]) -> T_fd:
        """Return the output of self.for_task(task), and also invalidate `self`.

        This is useful for more precisely expressing intent, if we don't intend to use
        `self` after getting the new FileDescriptor for the other task.

        This is also somewhat optimized relative to just calling self.for_task then
        calling self.invalidate; the latter call will have to be async, but this call
        doesn't have to be async, since we know we won't be invalidating the last handle.

        """
        new = self.for_task(task)
        self.valid = False
        handles = self._remove_from_tracking()
        if len(handles) == 0:
            raise Exception("We just made handle B from handle A, "
                            "so we know there are at least two handles; "
                            "but after removing handle A, there are no handles left. Huh?")
        return new

    def _get_global_handles(self) -> WeakSet[BaseFileDescriptor]:
        return self.task.fd_table.near_to_handles[self.near]

    def is_only_handle(self) -> bool:
        self._validate()
        return len(self._get_global_handles()) == 1

    def _remove_from_tracking(self) -> WeakSet[BaseFileDescriptor]:
        self.task.fd_handles.remove(self)
        handles = self._get_global_handles()
        handles.remove(self)
        return handles

    def __int__(self) -> int:
        return self.near.number

async def _close(sysif: SyscallInterface, fd: rsyscall.near.FileDescriptor) -> None:
    try:
        await sysif.syscall(SYS.close, fd)
    except OSError as e:
        e.filename = fd
        raise

class FDTable(rsyscall.far.FDTable):
    def __init__(self, creator_pid: int, parent: FDTable=None) -> None:
        super().__init__(creator_pid)
        self.near_to_handles: t.Dict[rsyscall.near.FileDescriptor, WeakSet[BaseFileDescriptor]] = {}
        self.tasks: WeakSet[FileDescriptorTask] = WeakSet([])
        if parent:
            self.inherited: WeakSet[BaseFileDescriptor] = WeakSet(
                itertools.chain(itertools.chain.from_iterable(parent.near_to_handles.values()),
                                parent.inherited)
            )
        else:
            self.inherited = WeakSet()

    def remove_inherited(self) -> None:
        self.inherited = WeakSet()

    def _get_task_in_table(self) -> t.Optional[FileDescriptorTask]:
        for task in list(self.tasks):
            if task.fd_table is not self:
                self.tasks.remove(task)
            elif task.manipulating_fd_table:
                # skip tasks currently changing fd table
                pass
            else:
                return task
        return None

    async def _close_fd(self, task: FileDescriptorTask, fd: rsyscall.near.FileDescriptor) -> None:
        try:
            # TODO we don't have to block here, we can just send off the close without waiting for it,
            # because you aren't supposed to retry close on error.
            # well, except when we actually want to give the user the chance to see the error from close.
            await _close(task.sysif, fd)
        except SyscallHangup:
            # closing the fd through this task went wrong
            # TODO we should mark this task as dead and fall back to later tasks in the list if
            # we fail due to a SyscallInterface-level error; that might happen if, say, this is
            # some decrepit task where we closed the syscallinterface but didn't exit the task.
            assert fd not in self.near_to_handles, f"fd {fd} was somehow reopened before it was actually closed"
            # put the fd back, some other task will close it
            self.near_to_handles[fd] = WeakSet()

    async def gc_using_task(self, task: FileDescriptorTask) -> None:
        gc.collect()
        async with trio.open_nursery() as nursery:
            # take a snapshot of near_to_handles so we can mutate it while iterating
            for fd, handles in list(self.near_to_handles.items()):
                if not handles:
                    # we immediately take responsibility for closing this fd, so our close
                    # attempts don't collide with others
                    del self.near_to_handles[fd]
                    logger.debug("gc for %s: starting close fd for %s", self, fd)
                    nursery.start_soon(self._close_fd, task, fd)

    async def run_gc(self) -> None:
        task = self._get_task_in_table()
        if task is not None:
            await self.gc_using_task(task)

class FileDescriptorTask(rsyscall.far.Task, t.Generic[T_fd]):
    def __init__(self,
                 sysif: rsyscall.near.SyscallInterface,
                 near_process: rsyscall.near.Process,
                 fd_table: FDTable,
                 address_space: rsyscall.far.AddressSpace,
                 pidns: rsyscall.far.PidNamespace,
    ) -> None:
        if not isinstance(fd_table, FDTable):
            raise Exception("fd_table", fd_table, "needs to be an", FDTable,
                            "to work with a", FileDescriptorTask)
        self.fd_table: FDTable
        super().__init__(sysif, near_process, fd_table, address_space, pidns)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.fd_handles: WeakSet[T_fd] = WeakSet()
        self.manipulating_fd_table = False
        self._add_to_active_fd_table_tasks()

    # for extensibility
    @abc.abstractmethod
    def _file_descriptor_constructor(self, fd: rsyscall.near.FileDescriptor) -> T_fd: ...

    def _make_fd_handle_from_near(self, fd: rsyscall.near.FileDescriptor) -> T_fd:
        if self.manipulating_fd_table:
            raise Exception("can't make a new FD handle while manipulating_fd_table==True")
        handle = self._file_descriptor_constructor(fd)
        logger.debug("%s: made handle %s from %s", self, handle, fd)
        self.fd_handles.add(handle)
        self.fd_table.near_to_handles.setdefault(fd, WeakSet()).add(handle)
        return handle

    def make_fd_handle(self, fd: rsyscall.near.FileDescriptor) -> T_fd:
        """Make a `FileDescriptor` referencing `fd`, which is currently unknown to rsyscall

        This is an unsafe operation: Since `rsyscall.near.FileDescriptor` is just a fancy `int`, we
        could use this to create handles for FDs which don't exist, or which are already being
        managed by others outside rsyscall. So use this with care.

        If this FD was inherited from a non-rsyscall parent process, you probably want to call
        `rsyscall.FileDescriptor.enable_cloexec` afterwards.

        """
        if fd in self.fd_table.near_to_handles:
            raise Exception("This fd is already known to us", fd)
        return self._make_fd_handle_from_near(fd)

    def inherit_fd(self, fd: BaseFileDescriptor) -> T_fd:
        """Make another FileDescriptor referencing `fd`, which we inherited, using this task for syscalls

        Whenever we call `clone` (without passing `CLONE.FILES`),
        the file descriptors in the parent process are copied to the child process;
        this is tracked in rsyscall, and we can call `inherit_fd` on any `FileDescriptor` in the parent process
        to get a handle for the copy of the file descriptor in the child.

        This same call also works if the two tasks are in the same file descriptor table, for convenience.

        """
        fd._validate()
        if fd.task.fd_table == self.fd_table:
            # let's allow "inheriting" fds that are already in.. the right fd table?
            # cuz... that way we can treat the case where we unshare and where we don't, identically
            # right? yes. it has to be the samed fd table, not the same task,
            # because the fds we're inheriting aren't in the same task.
            return self._make_fd_handle_from_near(fd.near)
        elif fd in self.fd_table.inherited:
            return self._make_fd_handle_from_near(fd.near)
        else:
            raise Exception("tried to inherit non-inherited fd", fd)

    def _add_to_active_fd_table_tasks(self) -> None:
        self.fd_table.tasks.add(self)

    def _make_fresh_fd_table(self) -> FDTable:
        """Make a new fd table that is a copy of the old one

        This is called by unshare_files, exec, and exit.

        """
        self.fd_table = FDTable(self.near_process.id)
        near_to_handles = self.fd_table.near_to_handles
        for handle in self.fd_handles:
            if handle.near not in near_to_handles:
                near_to_handles[handle.near] = WeakSet()
            near_to_handles[handle.near].add(handle)
        return self.fd_table

    async def run_fd_table_gc(self, use_self: bool=True) -> None:
        if use_self:
            await self.fd_table.gc_using_task(self)
        else:
            await self.fd_table.run_gc()

    async def unshare_files(self) -> None:
        """Unshare this task's file descriptor table.

        When such an unshare is done, the new file descriptor table may contain file
        descriptors which were copied from the old file descriptor table but are not now
        referenced by any `FileDescriptor`. Likewise, the old file descriptor table may
        contain file descriptors which are no longer referenced by any `FileDescriptor`,
        since the `FileDescriptor`s that referenced them were all for the task that unshared
        its table.  To remove such garbage, run_fd_table_gc is called for both the new and
        old fd tables after the unshare is complete.

        All the `FileDescriptor`s which access files through the task will remain valid.
        Linux will copy all the file descriptors from the old file descriptor table to the new file descriptor table,
        keeping them at the same numbers.
        The `FileDescriptor`s for the task will still be referencing the same file,
        through different file descriptors, which happen to have the same file descriptor *numbers*,
        in a new file descriptor table.
        Since the file descriptor numbers do not change, `near.FileDescriptor` will not change either,
        and no actual change is required in the `FileDescriptor`s.

        """
        if self.manipulating_fd_table:
            raise Exception("can't unshare_files while manipulating_fd_table==True")
        # do a GC now to improve efficiency when GCing both tables after the unshare
        gc.collect()
        old_fd_table = self.fd_table
        await old_fd_table.gc_using_task(self)
        self.manipulating_fd_table = True
        new_fd_table = self._make_fresh_fd_table()
        self._add_to_active_fd_table_tasks()
        # perform the actual unshare
        await _unshare(self.sysif, CLONE.FILES)
        # Each fd in the old table in the old table is also in the new table; this includes unwanted
        # fds that had handles in the old table and now don't have any handles.  Various race
        # conditions make garbage collecting those unwanted fds quite difficult, and ultimately
        # impossible.
        # The right way to close those unwanted fds is to, after the unshare, temporarily unset
        # CLOEXEC on the fds that we want to preserve, then close all CLOEXEC fds.  But
        # unfortunately there's no cheap way to close all CLOEXC fds.
        # So, instead, we just let the unwanted fds leak into the new table.  Almost all threads that call
        # unshare_files will call exec soon after, which will handle closing all CLOEXEC fds for us,
        # and so the leaked unwanted fds will be closed.
        # Concretely, we'll only create handles in the new table for the fds that this task owns, so
        # only those fds are involved in garbage collection; that handle-creation happened in
        # _make_fresh_fd_table.
        # TODO add a syscall to close all CLOEXEC fds, and call it here with appropriate setup
        self.manipulating_fd_table = False
        # We can only remove our handles from the handle lists after the unshare is done
        # and the fds are safely copied, because otherwise someone else running GC on the
        # old fd table would close our fds when they notice there are no more handles.
        for handle in self.fd_handles:
            old_fd_table.near_to_handles[handle.near].remove(handle)
        # GC the old fd table to delete any fds that are no longer referenced by our handles.
        await old_fd_table.run_gc()
        # GC the new table just to make sure that GC works; this should be a no-op, but we might
        # have messed something up.
        await new_fd_table.gc_using_task(self)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sched import CpuSet

class TestEventfd(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test(self) -> None:
        avail = await (await self.thr.task.sched_getaffinity(await self.thr.malloc(CpuSet))).read()
        with self.assertRaises(OSError, msg="calling setaffinity with an empty set should fail"):
            await self.thr.task.sched_setaffinity(await self.thr.ptr(CpuSet()))
        await self.thr.task.sched_setaffinity(await self.thr.ptr(CpuSet([list(avail)[0]])))
        await self.thr.task.sched_setaffinity(await self.thr.ptr(avail))
